# ==============================================================================
# Copyright (C) 2024 Shipei Xing <philipxsp@hotmail.com>
#
# Licensed under the Apache License 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at https://github.com/Philipbear/msbuddy/blob/main/LICENSE
# ==============================================================================
"""
File: ml.py
Author: Shipei Xing
Email: philipxsp@hotmail.com
GitHub: Philipbear
Description: machine learning functions: feature generation, prediction, etc.
             False discovery rate (FDR) estimation
"""
import math
import sys
import warnings

import numpy as np
from numba import njit
from scipy.stats import norm
from tqdm import tqdm

from msbuddy.query import common_nl_from_array, check_formula_existence

# ignore warnings
warnings.filterwarnings('ignore')


def _gen_arr_from_buddy_data(buddy_data) -> (np.array, np.array, np.array):
    """
    generate three arrays for ML
    :param buddy_data: List of MetaFeature objects
    :return: all_cand_form_arr, dbe_arr, mass_arr
    """
    all_cand_form_arr = np.empty((0, 12))  # formula array of all candidate formulas
    dbe_arr = np.array([])
    mass_arr = np.array([])

    for mf in buddy_data:
        if not mf.candidate_formula_list:
            continue
        # generate ML features for each candidate formula
        for cf in mf.candidate_formula_list:
            # add formula array to all_cand_form_arr
            all_cand_form_arr = np.append(all_cand_form_arr, [cf.formula.array], axis=0)
            dbe_arr = np.append(dbe_arr, cf.formula.dbe)
            mass_arr = np.append(mass_arr, cf.formula.mass)

    return all_cand_form_arr, dbe_arr, mass_arr


@njit
def _gen_form_feature(all_cf_arr, dbe_arr, mass_arr) -> np.array:
    """
    generate formula features for ML
    :param all_cand_form_arr: numpy array of all candidate formula arrays
    :param dbe_arr: numpy array of all candidate formula dbe
    :param mass_arr: numpy array of all candidate formula mass
    :return: numpy array of ML features
    """
    # calculate ML features
    ele_sum_1_arr = all_cf_arr[:, 2] + all_cf_arr[:, 3] + all_cf_arr[:, 4] + all_cf_arr[:, 5] + \
                    all_cf_arr[:, 6] + all_cf_arr[:, 8]
    ele_sum_2_arr = ele_sum_1_arr + all_cf_arr[:, 10] + all_cf_arr[:, 11]
    chon_arr = np.clip(ele_sum_2_arr, 0, 1)  # whether other eles other than C, H, O, N exist
    chonps_arr = np.clip(ele_sum_1_arr, 0, 1)  # whether other eles other than C, H, O, N, P, S exist

    hal_arr = np.sum(all_cf_arr[:, 2:6], axis=1)  # sum of halogen atoms
    ta_arr = np.sum(all_cf_arr, axis=1)  # total number of atoms
    f_exist_arr = np.clip(all_cf_arr[:, 4], 0, 1)  # whether F exists
    cl_exist_arr = np.clip(all_cf_arr[:, 3], 0, 1)  # whether Cl exists
    br_exist_arr = np.clip(all_cf_arr[:, 2], 0, 1)  # whether Br exists
    i_exist_arr = np.clip(all_cf_arr[:, 5], 0, 1)  # whether I exists
    hal_ele_type_arr = f_exist_arr + cl_exist_arr + br_exist_arr + i_exist_arr  # number of halogen elements
    hal_two = np.clip(hal_ele_type_arr - 1, 0, 1)  # whether more than one halogen element exists
    hal_three = np.clip(hal_ele_type_arr - 2, 0, 1)  # whether more than two halogen elements exist
    senior_1_1_arr = (6 * all_cf_arr[:, 11] + 5 * all_cf_arr[:, 10] + 4 * all_cf_arr[:, 0] + 3 * all_cf_arr[:, 7] +
                      2 * all_cf_arr[:, 9] + all_cf_arr[:, 1] + hal_arr)
    senior_1_2_arr = all_cf_arr[:, 7] + all_cf_arr[:, 10] + all_cf_arr[:, 1] + hal_arr

    # halogen to H ratio, fill 0 if H = 0
    hal_h_arr = np.zeros(len(hal_arr))
    # if H > 0
    h_bool_arr = all_cf_arr[:, 1] > 0
    hal_h_arr[h_bool_arr] = hal_arr[h_bool_arr] / all_cf_arr[h_bool_arr, 1]

    # O/P ratio, fill 1 if phosphorus = 0
    o_p_arr = np.ones(len(hal_arr))
    # if phosphorus > 0
    p_bool_arr = all_cf_arr[:, 10] > 0
    o_p_arr[p_bool_arr] = all_cf_arr[p_bool_arr, 9] / all_cf_arr[p_bool_arr, 10] / 3

    # generate output array
    out = np.empty((len(all_cf_arr), 26))
    # populate output array
    for i in range(len(all_cf_arr)):
        ta = ta_arr[i]
        chno = all_cf_arr[i, 0] + all_cf_arr[i, 1] + all_cf_arr[i, 7] + all_cf_arr[i, 9]
        # if C > 0
        if all_cf_arr[i, 0] > 0:
            out[i, :] = [1 - chon_arr[i], 1 - chonps_arr[i],
                         all_cf_arr[i, 0] / ta, all_cf_arr[i, 1] / ta,
                         all_cf_arr[i, 7] / ta,
                         all_cf_arr[i, 9] / ta, all_cf_arr[i, 10] / ta,
                         all_cf_arr[i, 11] / ta, chno / ta,
                         hal_arr[i] / ta, senior_1_1_arr[i], senior_1_2_arr[i], 2 * ta - 1, dbe_arr[i],
                         np.sqrt(dbe_arr[i] / mass_arr[i]), dbe_arr[i] / np.power(mass_arr[i] / 100, 2 / 3),
                         all_cf_arr[i, 1] / all_cf_arr[i, 0],
                         all_cf_arr[i, 7] / all_cf_arr[i, 0],
                         all_cf_arr[i, 9] / all_cf_arr[i, 0],
                         all_cf_arr[i, 10] / all_cf_arr[i, 0],
                         all_cf_arr[i, 11] / all_cf_arr[i, 0],
                         hal_arr[i] / all_cf_arr[i, 0],
                         hal_h_arr[i], o_p_arr[i], hal_two[i], hal_three[i]]
        else:
            out[i, :] = [1 - chon_arr[i], 1 - chonps_arr[i],
                         all_cf_arr[i, 0] / ta, all_cf_arr[i, 1] / ta,
                         all_cf_arr[i, 7] / ta,
                         all_cf_arr[i, 9] / ta, all_cf_arr[i, 10] / ta,
                         all_cf_arr[i, 11] / ta, chno / ta,
                         hal_arr[i] / ta, senior_1_1_arr[i], senior_1_2_arr[i], 2 * ta - 1, dbe_arr[i],
                         np.sqrt(dbe_arr[i] / mass_arr[i]), dbe_arr[i] / np.power(mass_arr[i] / 100, 2 / 3),
                         0, 0, 0, 0, 0, 0,
                         hal_h_arr[i], o_p_arr[i], hal_two[i], hal_three[i]]

    return out


def _fill_form_feature_arr_in_batch_data(batch_data, feature_arr) -> None:
    """
    fill ML features in batch data
    :param batch_data: List of MetaFeature objects
    :param feature_arr: 2D numpy array of ML features
    :return: None
    """
    cnt = 0
    for mf in batch_data:
        if not mf.candidate_formula_list:
            continue
        # fill in ML features for each candidate formula
        for cf in mf.candidate_formula_list:
            cf.formula_feature_array = feature_arr[cnt, :]
            cnt += 1
    return


def gen_ml_feature(meta_feature_list, ppm: bool, ms1_tol: float, ms2_tol: float, gd) -> np.array:
    """
    generate ML features for all metabolic features
    :param meta_feature_list: List of MetaFeature objects
    :param ppm: whether to use ppm error
    :param ms1_tol: m/z tolerance for MS1
    :param ms2_tol: m/z tolerance for MS2
    :param gd: global dependencies
    :return: numpy array of ML features
    """
    # generate feature array and group size array
    total_X_arr = np.array([])

    for mf in meta_feature_list:
        if not mf.candidate_formula_list:
            continue

        # generate ML features for each candidate formula
        for cf in mf.candidate_formula_list:
            # get ML features
            ml_feature_arr = gen_ml_feature_single(mf, cf, ppm, ms1_tol, ms2_tol, gd)

            # add to feature array
            if total_X_arr.size == 0:
                total_X_arr = ml_feature_arr
            else:
                total_X_arr = np.vstack((total_X_arr, ml_feature_arr))

    # ensure 2d array
    if len(total_X_arr.shape) == 1:
        total_X_arr = np.expand_dims(total_X_arr, axis=0)

    return total_X_arr


def gen_ml_feature_single(meta_feature, cand_form, ppm: bool, ms1_tol: float, ms2_tol: float, gd) -> np.array:
    """
    generate ML features for a single candidate formula
    :param meta_feature: MetaFeature object
    :param cand_form: CandidateFormula object
    :param ppm: whether to use ppm error
    :param ms1_tol: m/z tolerance for MS1
    :param ms2_tol: m/z tolerance for MS2
    :param gd: global dependencies
    :return: numpy array of ML features
    """

    this_adduct = meta_feature.adduct

    # mz error in ppm
    theo_mass = cand_form.charged_formula.mass / abs(this_adduct.charge)
    mz_error = (meta_feature.mz - theo_mass) / theo_mass * 1e6 if ppm else meta_feature.mz - theo_mass
    mz_error_log_p = _calc_log_p_norm(mz_error, ms1_tol / 3)

    # precursor charged formula
    pre_charged_arr = cand_form.charged_formula.array
    pre_dbe = cand_form.charged_formula.dbe
    pre_h2c = pre_charged_arr[1] / pre_charged_arr[0] if pre_charged_arr[0] > 0 else 0  # no carbon, assign 0

    # MS1 isotope similarity
    ms1_iso_sim = cand_form.ms1_isotope_similarity if cand_form.ms1_isotope_similarity else 0

    # MS/MS-related features
    ms2_feature_arr = _gen_ms2_feature(meta_feature, cand_form, pre_dbe, pre_h2c, ppm, ms2_tol, gd)

    # pos mode bool
    pos_mode = 1 if this_adduct.charge > 0 else 0

    # generate output array
    out = np.concatenate((np.array([ms1_iso_sim]), np.array([mz_error_log_p]), np.array([pos_mode]),  # 3
                          cand_form.formula_feature_array, ms2_feature_arr))  # 26 + 24

    return out


def _gen_ms2_feature(meta_feature, cand_form, pre_dbe: float, pre_h2c: float,
                     ppm: bool, ms2_tol: float, gd) -> np.array:
    """
    generate MS/MS-related features for a single candidate formula
    :param meta_feature: MetaFeature object
    :param cand_form: CandidateFormula object
    :param pre_dbe: precursor DBE
    :param pre_h2c: precursor H/C ratio
    :param ppm: whether to use ppm error
    :param ms2_tol: m/z tolerance for MS2
    :param gd: global dependencies
    :return: numpy array of MS/MS-related features
    """
    # MS2 explanation
    ms2_explanation = cand_form.ms2_raw_explanation

    # valid MS2 explanation
    if meta_feature.ms2_processed and ms2_explanation:
        # explained fragment ion
        exp_idx_arr = ms2_explanation.idx_array
        exp_int_arr = meta_feature.ms2_raw.int_array[exp_idx_arr]
        exp_mz_arr = meta_feature.ms2_raw.mz_array[exp_idx_arr]
        # valid fragment ion after MS2 processing
        valid_idx_arr = meta_feature.ms2_processed.idx_array
        valid_int_arr = meta_feature.ms2_raw.int_array[valid_idx_arr]

        # explained fragment ion count percentage
        exp_frag_cnt_pct = len(exp_idx_arr) / len(valid_idx_arr)

        # explained fragment ion intensity percentage
        exp_frag_int_pct = np.sum(exp_int_arr) / np.sum(valid_int_arr)

        frag_form_list = ms2_explanation.explanation_list  # list of fragment formulas, Formula objects

        pos_mode = meta_feature.adduct.pos_mode

        # check db existence of all explained fragments and neutral losses
        frag_db_existed, frag_common = np.array([], dtype=bool), np.array([], dtype=bool)
        nl_db_existed, nl_common = np.array([], dtype=bool), np.array([], dtype=bool)
        for f in frag_form_list:
            frag_db_bool, frag_common_bool = check_formula_existence(f.array, pos_mode, True, gd)
            frag_db_existed = np.append(frag_db_existed, frag_db_bool)
            frag_common = np.append(frag_common, frag_common_bool)
            nl_db_bool, nl_common_bool = check_formula_existence(cand_form.charged_formula.array - f.array,
                                                                 pos_mode, False, gd)
            nl_db_existed = np.append(nl_db_existed, nl_db_bool)
            nl_common = np.append(nl_common, nl_common_bool)

        # logical OR of fragment/nl
        fragnl_db_existed = np.logical_or(frag_db_existed, nl_db_existed)
        fragnl_common = np.logical_or(frag_common, nl_common)

        # explained and db existed fragment/nl ion count percentage
        exp_db_frag_cnt_pct = np.sum(frag_db_existed) / len(valid_idx_arr)
        exp_db_nl_cnt_pct = np.sum(nl_db_existed) / len(valid_idx_arr)
        exp_db_fragnl_cnt_pct = np.sum(fragnl_db_existed) / len(valid_idx_arr)

        # explained and db existed fragment/nl ion intensity percentage
        exp_db_frag_int_pct = np.sum(exp_int_arr[frag_db_existed]) / np.sum(valid_int_arr)
        exp_db_nl_int_pct = np.sum(exp_int_arr[nl_db_existed]) / np.sum(valid_int_arr)
        exp_db_fragnl_int_pct = np.sum(exp_int_arr[fragnl_db_existed]) / np.sum(valid_int_arr)

        # common fragment/nl ion count percentage
        exp_common_frag_cnt_pct = np.sum(frag_common) / len(valid_idx_arr)
        exp_common_nl_cnt_pct = np.sum(nl_common) / len(valid_idx_arr)
        exp_common_fragnl_cnt_pct = np.sum(fragnl_common) / len(valid_idx_arr)

        # common fragment/nl ion intensity percentage
        exp_common_frag_int_pct = np.sum(exp_int_arr[frag_common]) / np.sum(valid_int_arr)
        exp_common_nl_int_pct = np.sum(exp_int_arr[nl_common]) / np.sum(valid_int_arr)
        exp_common_fragnl_int_pct = np.sum(exp_int_arr[fragnl_common]) / np.sum(valid_int_arr)

        # subformula count: how many frags are subformula of other frags
        subform_score, subform_common_loss_score = _calc_subformula_score(frag_form_list, gd)

        # radical ion count percentage (out of all explained fragment ions)
        radical_cnt_pct = np.sum([1 for frag_form in frag_form_list if frag_form.dbe % 1 == 0]) / len(frag_form_list)

        # normalized explained intensity array
        normed_exp_int_arr = exp_int_arr / np.sum(exp_int_arr)

        # weighted average of fragment DBEs
        frag_dbe_wavg = np.sum(np.array([frag_form.dbe for frag_form in frag_form_list]) * normed_exp_int_arr)

        # weighted average of fragment H/C ratios
        frag_h2c_wavg = np.sum(np.array([frag_form.array[1] / frag_form.array[0] if frag_form.array[0] > 0 else pre_h2c
                                         for frag_form in frag_form_list]) * normed_exp_int_arr)

        # weighted average of fragment m/z ppm errors
        if ppm:
            frag_mz_err_wavg = np.sum(np.array([_calc_log_p_norm((frag_form.mass - mz) / frag_form.mass * 1e6,
                                                                 ms2_tol / 3)
                                                for frag_form, mz in
                                                zip(frag_form_list, exp_mz_arr)]) * normed_exp_int_arr)
        else:
            frag_mz_err_wavg = np.sum(np.array([_calc_log_p_norm(frag_form.mass - mz, ms2_tol / 3)
                                                for frag_form, mz in
                                                zip(frag_form_list, exp_mz_arr)]) * normed_exp_int_arr)

        # weighted average of fragment-nl DBE difference
        frag_nl_dbe_diff_wavg = np.sum(np.array([frag_form.dbe - (pre_dbe - frag_form.dbe + 1)
                                                 for frag_form in frag_form_list]) * normed_exp_int_arr)

        out_arr = np.array([exp_frag_cnt_pct, exp_frag_int_pct, exp_db_frag_cnt_pct, exp_db_nl_cnt_pct,
                            exp_db_fragnl_cnt_pct, exp_db_frag_int_pct, exp_db_nl_int_pct, exp_db_fragnl_int_pct,
                            exp_common_frag_cnt_pct, exp_common_nl_cnt_pct, exp_common_fragnl_cnt_pct,
                            exp_common_frag_int_pct, exp_common_nl_int_pct, exp_common_fragnl_int_pct,
                            subform_score, subform_common_loss_score,
                            radical_cnt_pct, frag_dbe_wavg, frag_h2c_wavg, frag_mz_err_wavg, frag_nl_dbe_diff_wavg,
                            len(valid_idx_arr), math.sqrt(exp_frag_cnt_pct), math.sqrt(exp_frag_int_pct)])
    else:
        out_arr = np.array([0] * 24)

    return out_arr


def _calc_subformula_score(frag_form_arr, gd) -> (float, float):
    """
    calculate how many formulas are subformula of other formulas, generate corresponding scores
    :param frag_form_arr: list of Formula objects
    :return: subformula count, subformula count with common loss
    """
    # 0 or 1 frag explanation
    if len(frag_form_arr) <= 1:
        return 0, 0

    exp_frag_cnt = 0  # explained fragment count, except for isotope peaks
    # loop through all fragment formulas, stack formula arrays that are not isotope peaks
    all_frag_arr = frag_form_arr[0].array
    for i in range(1, len(frag_form_arr)):
        if frag_form_arr[i].isotope > 0:
            continue
        all_frag_arr = np.vstack((all_frag_arr, frag_form_arr[i].array))
        exp_frag_cnt += 1

    if exp_frag_cnt <= 1:
        return 0, 0

    # subformula check & subformula common loss check
    subform_cnt, subform_common_loss_cnt = _subformula_check(all_frag_arr, gd['common_loss_db'])

    # generate scores, normalized by the number of all possible combinations
    subform_score = 2 * subform_cnt / (exp_frag_cnt * (exp_frag_cnt - 1))
    subform_common_loss_score = 2 * subform_common_loss_cnt / (exp_frag_cnt * (exp_frag_cnt - 1))

    return subform_score, subform_common_loss_score


@njit
def _subformula_check(all_frag_arr: np.array, nl_db: np.array):
    """
    check if a formula is subformula of another formula
    :param all_frag_arr: array of fragment formulas, array of formula arrays
    :param nl_db: common loss database
    :return: subformula count, subformula count with common loss
    """
    subform_cnt = 0
    subform_common_loss_cnt = 0

    for i in range(len(all_frag_arr)):
        for j in range(i + 1, len(all_frag_arr)):
            delta_arr = all_frag_arr[j] - all_frag_arr[i]
            # check if subformula; check only C > 0, N > 0
            if np.all(delta_arr >= 0) and np.sum(delta_arr[1:]) > 0 and (np.sum(delta_arr) - delta_arr[7]) > 0:
                subform_cnt += 1
                # check if there is common loss
                if common_nl_from_array(delta_arr, nl_db):
                    subform_common_loss_cnt += 1

    return subform_cnt, subform_common_loss_cnt


def _calc_log_p_norm(arr: np.array, sigma: float) -> np.array:
    """
    calculate log(p) for each element in an array, where p is the probability of a normal distribution
    :param arr: numpy array
    :param sigma: sigma for normal distribution
    :return: numpy array
    """
    arr_norm_p = norm.cdf(arr, loc=0, scale=sigma)
    return _calc_log_p_norm_helper(arr_norm_p)


@njit
def _calc_log_p_norm_helper(arr_norm_p) -> np.array:
    """
    calculate log(p) for a single element, where p is the probability of a normal distribution
    :param arr_norm_p: array of probabilities
    :return: numpy array of log(p)
    """
    arr_norm_p = 1 - arr_norm_p if arr_norm_p > 0.5 else arr_norm_p
    log_p = np.log(arr_norm_p * 2)
    # clip to the range of [-2, 0]
    log_p = -2 if log_p < -2 else log_p

    return log_p


def _predict_ml(meta_feature_list, group_no: int, ppm: bool, ms1_tol: float, ms2_tol: float, gd) -> np.array:
    """
    ml prediction
    :param meta_feature_list: List of MetaFeature objects
    :param group_no: group number; 0: ms1 ms2; 1: ms1 noms2; 2: noms1 ms2; 3: noms1 noms2
    :param ppm: whether to use ppm error
    :param ms1_tol: m/z tolerance for MS1
    :param ms2_tol: m/z tolerance for MS2
    :param gd: global dependencies
    :return: numpy array of prediction results
    """
    # generate feature array
    X_arr = gen_ml_feature(meta_feature_list, ppm, ms1_tol, ms2_tol, gd)

    if X_arr.size == 0:
        return np.array([])

    # load model
    if group_no == 0:
        model = gd['model_ms1_ms2']
    elif group_no == 1:
        model = gd['model_ms1_noms2']
        X_arr = X_arr[:, :-24]  # remove MS2-related features
    elif group_no == 2:
        model = gd['model_noms1_ms2']
        X_arr = X_arr[:, 1:]  # remove MS1 isotope similarity
    else:
        model = gd['model_noms1_noms2']
        X_arr = X_arr[:, 1:]  # remove MS1 isotope similarity
        X_arr = X_arr[:, :-24]  # remove MS2-related features

    # predict formula probability
    score_arr = model.predict(X_arr)
    return score_arr


def predict_formula_probability(buddy_data, batch_start_idx: int, batch_end_idx: int, config, gd):
    """
    predict formula probability
    :param buddy_data: buddy data
    :param batch_start_idx: batch start index
    :param batch_end_idx: batch end index
    :param config: config object
    :param gd: global dependencies
    :return: fill in estimated_prob in candidate formula objects
    """
    ppm = config.ppm
    ms1_tol = config.ms1_tol
    ms2_tol = config.ms2_tol

    # batch data
    batch_data = buddy_data[batch_start_idx:batch_end_idx]

    # generate three arrays from buddy data
    cand_form_arr, dbe_arr, mass_arr = _gen_arr_from_buddy_data(batch_data)

    # if no candidate formula, return
    if cand_form_arr.size == 0:
        return

    # generate ML feature array
    feature_arr = _gen_form_feature(cand_form_arr, dbe_arr, mass_arr)

    # fill in batch_data
    _fill_form_feature_arr_in_batch_data(batch_data, feature_arr)

    # split buddy data into 4 groups and store their indices in a dictionary
    group_dict = dict()
    for i in range(4):
        group_dict[i] = []

    for i, meta_feature in enumerate(batch_data):
        if not meta_feature.candidate_formula_list:
            continue
        if meta_feature.ms1_raw:
            if meta_feature.ms2_raw:
                group_dict[0].append(i)
            else:
                group_dict[1].append(i)
        else:
            if meta_feature.ms2_raw:
                group_dict[2].append(i)
            else:
                group_dict[3].append(i)

    # predict formula probability
    for i in range(4):
        if not group_dict[i]:
            continue
        # predict formula probability
        prob_arr = _predict_ml([batch_data[j] for j in group_dict[i]], i, ppm, ms1_tol, ms2_tol, gd)
        # Platt calibration
        prob_arr = _platt_calibrated_probability(prob_arr, gd['platt_a_' + str(i)], gd['platt_b_' + str(i)])
        # add prediction results to candidate formula objects in the list
        cnt = 0
        for j in group_dict[i]:
            for candidate_formula in batch_data[j].candidate_formula_list:
                candidate_formula.estimated_prob = prob_arr[cnt]
                cnt += 1

    # update buddy data
    buddy_data[batch_start_idx:batch_end_idx] = batch_data
    return


def _platt_calibrated_probability(score, a, b):
    """
    Platt calibration for FDR estimation
    :param score: predicted scores
    :param a: coefficient of the sigmoid function
    :param b: intercept of the sigmoid function
    :return: calibrated probability
    """
    probability = 1 / (1 + np.exp(-(a * score + b)))
    return probability


def calc_fdr(buddy_data, batch_start_idx: int, batch_end_idx: int):
    """
    calculate FDR for candidate formulas
    :param buddy_data: buddy data
    :param batch_start_idx: batch start index
    :param batch_end_idx: batch end index
    :return: fill in FDR in candidate formula objects
    """
    batch_data = buddy_data[batch_start_idx:batch_end_idx]

    # sort candidate formula list for each metabolic feature
    for meta_feature in tqdm(batch_data, desc="FDR calculation: ", file=sys.stdout, colour="green"):
        if not meta_feature.candidate_formula_list:
            continue
        # sort candidate formula list by estimated probability, in descending order
        meta_feature.candidate_formula_list.sort(key=lambda x: x.estimated_prob, reverse=True)

        # sum of estimated probabilities
        prob_sum = np.sum([cand_form.estimated_prob for cand_form in meta_feature.candidate_formula_list])

        # calculate normed estimated prob and FDR considering all candidate formulas
        sum_normed_estimated_prob = 0
        for i, cand_form in enumerate(meta_feature.candidate_formula_list):
            this_normed_estimated_prob = cand_form.estimated_prob / prob_sum
            sum_normed_estimated_prob += this_normed_estimated_prob

            cand_form.normed_estimated_prob = this_normed_estimated_prob
            cand_form.estimated_fdr = 1 - (sum_normed_estimated_prob / (i + 1))

        # if meta_feature.candidate_formula_list[0].estimated_prob > 0.1:
        #     # calculate normed estimated prob and FDR considering all candidate formulas
        #     sum_normed_estimated_prob = 0
        #     for i, cand_form in enumerate(meta_feature.candidate_formula_list):
        #         this_normed_estimated_prob = cand_form.estimated_prob / prob_sum
        #         sum_normed_estimated_prob += this_normed_estimated_prob
        #
        #         cand_form.normed_estimated_prob = this_normed_estimated_prob
        #         cand_form.estimated_fdr = 1 - (sum_normed_estimated_prob / (i + 1))
        # else:
        #     # scale estimated prob using softmax, to reduce the effect of very small probs
        #     prob_sum = np.sum(
        #         [np.exp(cand_form.estimated_prob) for cand_form in meta_feature.candidate_formula_list])
        #     sum_normed_estimated_prob = 0
        #     for i, cand_form in enumerate(meta_feature.candidate_formula_list):
        #         this_normed_estimated_prob = np.exp(cand_form.estimated_prob) / prob_sum
        #         sum_normed_estimated_prob += this_normed_estimated_prob
        #
        #         cand_form.normed_estimated_prob = this_normed_estimated_prob
        #         cand_form.estimated_fdr = 1 - (sum_normed_estimated_prob / (i + 1))

    # update back
    buddy_data[batch_start_idx:batch_end_idx] = batch_data
    return



# ==============================================================================
# Copyright (C) 2024 Shipei Xing <philipxsp@hotmail.com>
#
# Licensed under the Apache License 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at https://github.com/Philipbear/msbuddy/blob/main/LICENSE
# ==============================================================================
"""
File: utils.py
Author: Shipei Xing
Email: philipxsp@hotmail.com
GitHub: Philipbear
Description: main util functions for msbuddy
"""

from typing import List

import numpy as np
from chemparse import parse_formula
from numba import njit

# define alphabet
alphabet = ["C", "H", "Br", "Cl", "F", "I", "K", "N", "Na", "O", "P", "S"]


class FormulaResult:
    """
    FormulaResult class, for mass_to_formula function
    """

    def __init__(self, formula: str, mass: float, t_mass: float):
        """
        Constructor
        :param formula: formula string
        :param mass: exact mass of the formula
        :param t_mass: target mass
        """
        self.formula = formula
        self.mass_error = mass - t_mass
        self.mass_error_ppm = self.mass_error / t_mass * 1e6


class SubformulaResult:
    """
    SubformulaResult class, for assign_subformula function
    """

    def __init__(self, idx: int, subform_list: List[FormulaResult]):
        """
        Constructor
        :param idx: index of the fragment ion
        :param subform_list: list of FormulaResult objects
        """
        self.idx = idx
        self.subform_list = subform_list


def read_formula(form: str) -> np.array:
    """
    read neutral formula string and return a 12-dim array, return None if invalid
    :param form: formula string in neutral form, e.g., C6H5NO2
    :return: 12-dim array
    """

    # parse formula string
    try:
        parsed = parse_formula(form)
    except:
        return None

    # check whether contains elements not in alphabet
    for element in parsed.keys():
        if element not in alphabet:
            return None

    # convert to 12-dim array
    array = np.zeros(12, dtype=np.int16)
    for i, element in enumerate(alphabet):
        if element in parsed.keys():
            array[i] = parsed[element]
    return array


def enumerate_subform_arr(form_arr: List[int]) -> np.ndarray:
    """
    enumerate all possible sub-formula arrays of a given formula array
    :param form_arr: a list-like object
    :return: a 2-dim numpy array, with each row being a sub-formula array
    """

    form_arr = np.array(form_arr, dtype=np.int16)

    # enumerate all possible sub-formula arrays
    return enumerate_subformula(form_arr)


@njit
def enumerate_subformula(pre_charged_arr: np.array) -> np.array:
    """
    Enumerate all subformulas of a candidate formula. (Numba version)
    :param pre_charged_arr: precursor charged array
    :return: 2D array, each row is a subformula array
    """
    n = len(pre_charged_arr)
    total_subform_cnt = np.prod(pre_charged_arr + 1)

    subform_arr = np.zeros((total_subform_cnt, n), dtype=np.int16)
    tempSize = 1

    for i in range(n):
        count = pre_charged_arr[i]
        repeatSize = tempSize
        tempSize *= (count + 1)

        pattern = np.arange(count + 1)

        repeated_pattern = np.empty(repeatSize * len(pattern), dtype=np.int16)
        for j in range(len(pattern)):
            repeated_pattern[j * repeatSize: (j + 1) * repeatSize] = pattern[j]

        full_repeats = total_subform_cnt // len(repeated_pattern)

        for j in range(full_repeats):
            start_idx = j * len(repeated_pattern)
            end_idx = (j + 1) * len(repeated_pattern)
            subform_arr[start_idx:end_idx, i] = repeated_pattern

    return subform_arr


# for numba
alphabet_np = np.array(
    [ord(char) for word in ["C", "H", "Br", "Cl", "F", "I", "K", "N", "Na", "O", "P", "S"] for char in word],
    dtype=np.int16)
word_lengths = np.array([len(word) for word in ["C", "H", "Br", "Cl", "F", "I", "K", "N", "Na", "O", "P", "S"]],
                        dtype=np.int16)


@njit
def _form_arr_to_str(form_arr: np.array):
    """
    Inner func: convert formula array to string. (Numba version)
    :param form_arr: formula array
    :return: formula_list
    """
    formula_list = []
    idx = 0
    for i in range(len(word_lengths)):
        if form_arr[i]:
            for _ in range(word_lengths[i]):
                formula_list.append(alphabet_np[idx])
                idx += 1
            if form_arr[i] > 1:
                for digit in str(form_arr[i]):
                    formula_list.append(ord(digit))
        else:
            idx += word_lengths[i]
    return formula_list


def _ascii_to_str(ascii_arr) -> str:
    """
    Convert ASCII integer array to string
    :param ascii_arr: ASCII array
    :return: string
    """
    return ''.join(chr(i) for i in ascii_arr)


def form_arr_to_str(form_arr) -> str:
    """
    Convert formula array to string. (Numba version)
    :param form_arr: formula np.array
    :return: formula string
    """
    return _ascii_to_str(_form_arr_to_str(form_arr))


def _read_formula_str(x: str) -> dict:
    """
    Read formula string and return a dictionary
    :param x: formula string
    :return: formula dictionary
    """
    # if x starts with a number
    m = 1
    idx = 0
    while x[idx].isdigit():
        idx += 1
    if idx > 0:
        m = int(x[:idx])
    x = x[idx:]
    # parse formula
    form = parse_formula(x)
    # multiply by m
    for k, v in form.items():
        form[k] = v * m
    return form


def read_formula_str(x: str) -> dict:
    """
    Read formula string and return a dictionary
    :param x: formula string
    :return: formula dictionary
    """
    # if x contains '.', read two parts
    if '.' in x:
        x1 = x.split('.')[0]
        x2 = x.split('.')[1]
        form1 = _read_formula_str(x1)  # this is a dict
        form2 = _read_formula_str(x2)
        # merge two dicts, add values with same keys
        for k, v in form2.items():
            if k in form1.keys():
                form1[k] += v
            else:
                # add new key
                form1[k] = v
        return form1
    else:
        return _read_formula_str(x)


def add_formula_str(x: str, y: str) -> dict:
    """
    Add two formula strings
    :param x: formula string x
    :param y: formula string y
    :return: formula dictionary
    """
    form1 = read_formula_str(x)
    form2 = read_formula_str(y)
    # merge two dicts, add values with same keys
    for k, v in form2.items():
        if k in form1.keys():
            form1[k] += v
        else:
            # add new key
            form1[k] = v
    return form1



# ==============================================================================
# Copyright (C) 2024 Shipei Xing <philipxsp@hotmail.com>
#
# Licensed under the Apache License 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at https://github.com/Philipbear/msbuddy/blob/main/LICENSE
# ==============================================================================
"""
File: cand.py
Author: Shipei Xing
Email: philipxsp@hotmail.com
GitHub: Philipbear
Description: generate candidate formula space for a metabolic feature; mass search and bottom-up MS/MS interrogation
"""

from typing import Union, List, Tuple

import numpy as np
from brainpy import isotopic_variants
from numba import njit

from msbuddy.base import Formula, CandidateFormula, MS2Explanation, MetaFeature, check_adduct, Adduct
from msbuddy.ml import _calc_log_p_norm
from msbuddy.query import check_common_frag, check_common_nl, query_precursor_mass, query_fragnl_mass
from msbuddy.utils import form_arr_to_str, enumerate_subformula, read_formula, SubformulaResult, FormulaResult


class FragExplanation:
    """
    FragExplanation is a class for storing all potential fragment/nl explanations for a single MS/MS peak.
    It contains a list of fragment formulas, neutral loss formulas, and the index of the fragment.
    """

    def __init__(self, idx: int, frag: Formula, nl: Formula):
        self.idx = idx  # raw MS2 peak index
        self.frag_list = [frag]  # List[Formula]
        self.nl_list = [nl]  # List[Formula]
        self.optim_nl = None
        self.optim_frag = None

    def direct_assign_optim(self):
        """
        Directly assign the first explanation as the optimized explanation.
        """
        self.optim_frag = self.frag_list[0]
        self.optim_nl = self.nl_list[0]

    def add_frag_nl(self, frag: Formula, nl: Formula):
        self.frag_list.append(frag)
        self.nl_list.append(nl)

    def __len__(self):
        return len(self.frag_list)

    def refine_explanation_v1(self, raw_ms2_mz_arr: np.array, gd):
        """
        Refine the MS2 explanation by selecting the most reasonable explanation.
        :param raw_ms2_mz_arr: raw MS2 m/z array
        :param gd: global dictionary
        :return: fill in self.optim_frag, self.optim_nl
        """
        if len(self) == 1:
            self.optim_frag = self.frag_list[0]
            self.optim_nl = self.nl_list[0]
            return

        # if multiple explanations, select the most reasonable one
        # 1. check common neutral loss/fragment
        # 2. select the closest to the raw MS2 m/z

        # either common frag or common nl is True
        common_bool = [check_common_frag(frag, gd) or check_common_nl(nl, gd) for frag, nl in
                       zip(self.frag_list, self.nl_list)]
        # if only one common frag/nl, select it
        if sum(common_bool) == 1:
            idx = common_bool.index(True)
            self.optim_frag = self.frag_list[idx]
            self.optim_nl = self.nl_list[idx]
            return
        # if multiple common frag/nl, select the mz closest one
        elif sum(common_bool) > 1:
            # get the mass-closest one
            idx = np.argmin(np.abs(raw_ms2_mz_arr[self.idx] - np.array([f.mass for i, f in enumerate(self.frag_list)
                                                                        if common_bool[i]])))
            self.optim_frag = [f for i, f in enumerate(self.frag_list) if common_bool[i]][idx]
            self.optim_nl = [f for i, f in enumerate(self.nl_list) if common_bool[i]][idx]
            return
        # if no common frag/nl, select the closest one
        else:
            idx = np.argmin(np.abs(raw_ms2_mz_arr[self.idx] - np.array([f.mass for f in self.frag_list])))
            self.optim_frag = self.frag_list[idx]
            self.optim_nl = self.nl_list[idx]
            return

    def refine_explanation(self, raw_ms2_mz_arr: np.array):
        """
        Refine the MS2 explanation by selecting the most reasonable explanation.
        :param raw_ms2_mz_arr: raw MS2 m/z array
        :return: fill in self.optim_frag, self.optim_nl
        """
        if len(self) == 1:
            self.optim_frag = self.frag_list[0]
            self.optim_nl = self.nl_list[0]
            return

        # if multiple explanations, select the closest to the raw MS2 m/z
        idx = _find_closest_mass_idx(raw_ms2_mz_arr[self.idx], np.array([f.mass for f in self.frag_list]))
        # idx = np.argmin(np.abs(raw_ms2_mz_arr[self.idx] - np.array([f.mass for f in self.frag_list])))
        self.optim_frag = self.frag_list[idx]
        self.optim_nl = self.nl_list[idx]
        return


class CandidateSpace:
    """
    CandidateSpace is a class for bottom-up MS/MS interrogation.
    It contains a precursor candidate and a list of FragExplanations.
    """

    def __init__(self, pre_neutral_array: np.array, pre_charged_array: np.array,
                 pre_neutral_form_str: Union[str, None] = None,
                 exp_frag_sum_int: Union[float, None] = None,
                 exp_frag_idx_ls: Union[List[int], None] = None,
                 frag_exp_ls: Union[List[FragExplanation], None] = None):
        self.pre_neutral_array = np.int16(pre_neutral_array)  # precursor neutral array
        self.pre_charged_array = np.int16(pre_charged_array)  # used for ms2 global optim.
        self.pre_neutral_form_str = pre_neutral_form_str
        self.exp_frag_sum_int = exp_frag_sum_int  # during candidate space generation
        self.exp_frag_idx_ls = exp_frag_idx_ls  # during candidate space generation
        self.neutral_mass = float(np.sum(pre_neutral_array * Formula.mass_arr))
        self.frag_exp_list = frag_exp_ls  # during subformula assignment, List[FragExplanation]

    def add_frag_exp(self, frag_exp: FragExplanation):
        self.frag_exp_list.append(frag_exp)

    def refine_explanation(self, meta_feature: MetaFeature,
                           ms2_iso_tol: float) -> CandidateFormula:
        """
        Refine the MS2 explanation by selecting the most reasonable explanation.
        Explain other fragments as isotope peaks.
        Convert into a CandidateFormula.
        :param meta_feature: MetaFeature object
        :param ms2_iso_tol: MS2 tolerance for isotope peaks, in Da
        :return: CandidateFormula
        """
        ms2_raw = meta_feature.ms2_raw
        ms2_processed = meta_feature.ms2_processed
        # for each frag_exp, refine the explanation, select the most reasonable frag/nl
        for frag_exp in self.frag_exp_list:
            frag_exp.refine_explanation(ms2_raw.mz_array)

        # consider to further explain other fragments as isotope peaks
        explained_idx = [f.idx for f in self.frag_exp_list]
        for m, exp_idx in enumerate(explained_idx):
            # last peak, skip
            if m + 1 == len(explained_idx):
                continue
            # idx of next exp peak
            next_exp_idx = explained_idx[m + 1]
            # if the next peak is already explained
            if next_exp_idx in explained_idx:
                continue
            # if this idx is not in idx_array of ms2_processed, skip
            if next_exp_idx not in ms2_processed.idx_array:
                continue
            # if the next peak is close enough, add it as an isotope peak
            if (ms2_raw.mz_array[next_exp_idx] - ms2_raw.mz_array[exp_idx] - 1.003355) <= ms2_iso_tol and \
                    ms2_raw.int_array[next_exp_idx] <= ms2_raw.int_array[exp_idx]:
                # if the next peak is close enough, add it as an isotope peak
                this_frag = self.frag_exp_list[m].optim_frag
                this_nl = self.frag_exp_list[m].optim_nl
                # add a new FragExplanation
                new_frag = Formula(this_frag.array, this_frag.charge, this_frag.mass + 1.003355, 1)
                # for iso peak, the neutral loss is actually the same as the previous one (M+0)
                # but we denote it as '-1' isotope peak, so that frag mass + nl mass is still precursor mass
                new_nl = Formula(this_nl.array, 0, this_nl.mass - 1.003355, -1)
                new_frag_exp = FragExplanation(next_exp_idx, new_frag, new_nl)
                new_frag_exp.direct_assign_optim()
                self.frag_exp_list.append(new_frag_exp)

        # sort the frag_exp_list by idx
        self.frag_exp_list = sorted(self.frag_exp_list, key=lambda x: x.idx)

        # convert into a CandidateFormula
        # construct MS2Explanation first
        ms2_raw_exp = MS2Explanation(idx_array=np.array([f.idx for f in self.frag_exp_list], dtype=np.int16),
                                     explanation_list=[f.optim_frag for f in self.frag_exp_list])

        return CandidateFormula(formula=Formula(self.pre_neutral_array, 0, self.neutral_mass),
                                charged_formula=Formula(self.pre_charged_array, meta_feature.adduct.charge),
                                ms2_raw_explanation=ms2_raw_exp)


def gen_candidate_formula(mf: MetaFeature, ppm: bool, ms1_tol: float, ms2_tol: float,
                          db_mode: int, ele_lower_limit: np.array, ele_upper_limit: np.array,
                          max_isotope_cnt: int, gd: dict) -> MetaFeature:
    """
    Generate candidate formulas for a metabolic feature.
    :param mf: MetaFeature object
    :param ppm: whether to use ppm as the unit of tolerance
    :param ms1_tol: mz tolerance for precursor ion
    :param ms2_tol: mz tolerance for fragment ions / neutral losses
    :param db_mode: database mode (int, 0: basic; 1: halogen)
    :param ele_lower_limit: lower limit of each element
    :param ele_upper_limit: upper limit of each element
    :param max_isotope_cnt: maximum isotope count, used for MS1 isotope pattern matching
    :param gd: global dictionary
    :return: fill in list of candidate formulas (CandidateFormula) in metaFeature
    """

    # if MS2 data missing or non-singly charged species, query precursor mass directly
    if not mf.ms2_processed or abs(mf.adduct.charge) > 1:
        cf_list, _ = _gen_candidate_formula_from_mz(mf, ppm, ms1_tol,
                                                    ele_lower_limit, ele_upper_limit, db_mode, gd)

    else:
        # if MS2 data available, generate candidate space with MS2 data
        ms2_cand_form_ls, ms2_cand_form_str_ls = _gen_candidate_formula_from_ms2(mf, ppm, ms1_tol, ms2_tol,
                                                                                 ele_lower_limit,
                                                                                 ele_upper_limit, db_mode, gd)

        # query precursor mass, for fill in db_existed
        ms1_cand_form_ls, ms1_cand_form_str_ls = _gen_candidate_formula_from_mz(mf, ppm, ms1_tol, ele_lower_limit,
                                                                                ele_upper_limit, db_mode, gd)

        cf_list = _merge_cand_form_list(ms1_cand_form_ls, ms2_cand_form_ls,
                                        ms1_cand_form_str_ls, ms2_cand_form_str_ls)

        # if len(mf.ms2_processed) <= 5:
        #     # merge candidate formulas from ms1 and ms2
        #     cf_list = _merge_cand_form_list(ms1_cand_form_ls, ms2_cand_form_ls,
        #                                     ms1_cand_form_str_ls, ms2_cand_form_str_ls)
        # else:
        #     # fill in db_existed
        #     cf_list = _fill_in_db_existence(ms1_cand_form_ls, ms2_cand_form_ls,
        #                                     ms1_cand_form_str_ls, ms2_cand_form_str_ls)

    # calculate mz error
    cf_list = _calc_mz_error(cf_list, mf.mz, ppm)

    # if MS1 isotope data is available and >1 iso peaks, calculate isotope similarity
    ms1_iso_available = mf.ms1_processed and len(mf.ms1_processed) > 1
    if ms1_iso_available:
        for k, cf in enumerate(cf_list):
            cf_list[k].ms1_isotope_similarity = _calc_ms1_iso_sim(cf, mf, max_isotope_cnt)

    ms2_available = mf.ms2_processed

    # retain top candidate formulas
    mf.candidate_formula_list = _retain_top_cand_form(cf_list, ms1_tol, ms1_iso_available, ms2_available, 350)

    return mf


def _calc_mz_error(cf_list: List[CandidateFormula], precursor_mz: float, ppm: bool) -> List[CandidateFormula]:
    """
    Calculate mz error for each candidate formula.
    :param cf_list: list of candidate formulas
    :param precursor_mz: precursor mz
    :param ppm: whether to use ppm as the unit of tolerance
    :return: list of candidate formulas
    """
    if ppm:
        for k, cf in enumerate(cf_list):
            cf_list[k].mz_error = (cf.charged_formula.mass - precursor_mz) / precursor_mz * 1e6
    else:
        for k, cf in enumerate(cf_list):
            cf_list[k].mz_error = cf.charged_formula.mass - precursor_mz
    return cf_list


@njit
def _find_closest_mass_idx(mass: float, mass_arr: np.array) -> int:
    """
    find the closest mass in mass_arr to mass
    :param mass: float
    :param mass_arr: mass array
    :return: index of the closest mass
    """
    idx = np.argmin(np.abs(mass - mass_arr))
    return idx


def calc_isotope_pattern(formula: Formula, iso_peaks: Union[int, None] = 4) -> np.array:
    """
    calculate isotope pattern of a neutral formula with a given adduct
    :param formula: Formula object
    :param iso_peaks: number of isotope peaks to calculate
    :return: intensity array of isotope pattern
    """
    # mapping to a dictionary
    arr_dict = {}
    for i, element in enumerate(Formula.alphabet):
        arr_dict[element] = formula.array[i]

    # calculate isotope pattern
    isotope_pattern = isotopic_variants(arr_dict, npeaks=iso_peaks)
    int_arr = np.array([iso.intensity for iso in isotope_pattern], dtype=np.float32)

    return int_arr


@njit
def calc_isotope_similarity(int_arr_x, int_arr_y, iso_num: int) -> float:
    """
    calculate isotope similarity between two ms1 isotope patterns
    :param int_arr_x: intensity array of theoretical isotope pattern
    :param int_arr_y: intensity array of experimental isotope pattern
    :param iso_num: number of isotope peaks to calculate
    :return: isotope similarity
    """
    min_len = min(len(int_arr_x), iso_num)
    int_arr_x = int_arr_x[:min_len]  # theoretical isotope pattern
    if len(int_arr_y) > min_len:  # experimental isotope pattern
        int_arr_y = int_arr_y[:min_len]
    if len(int_arr_y) < min_len:
        int_arr_y = np.append(int_arr_y, np.zeros(min_len - len(int_arr_y), dtype=np.float32))

    # normalize
    int_arr_x = int_arr_x / np.sum(int_arr_x, dtype=np.float32)
    int_arr_y = int_arr_y / np.sum(int_arr_y, dtype=np.float32)

    # calculate the similarity
    int_diff_arr = np.abs(int_arr_y - int_arr_x)
    sim_score = 1 - np.sum(int_diff_arr)

    return sim_score


@njit
def _element_check(form_array: np.array, lower_limit: np.array, upper_limit: np.array) -> bool:
    """
    check whether a formula satisfies the element restriction
    :param form_array: 12-dim array
    :param lower_limit: 12-dim array
    :param upper_limit: 12-dim array
    :return: True if satisfies, False otherwise
    """
    if np.any(form_array < lower_limit) or np.any(form_array > upper_limit):
        return False
    return True


@njit
def _senior_rules(form: np.array) -> bool:
    """
    check whether a formula satisfies the senior rules
    :param form: 12-dim array
    :return: True if satisfies, False otherwise
    """
    # ["C", "H", "Br", "Cl", "F", "I", "K", "N", "Na", "O", "P", "S"]
    # int senior_1_1 = 6 * s + 5 * p + 4 * c + 3 * n + 2 * o + h + f + cl + br + i + na + k
    # int senior_1_2 = p + n + h + f + cl + br + i + na + k
    # int senior_2 = c + h + n + o + p + f + cl + br + i + s + na + k

    senior_1_1 = (6 * form[11] + 5 * form[10] + 4 * form[0] + 3 * form[7] + 2 * form[9] + form[1] + form[4] +
                  form[3] + form[2] + form[5] + form[8] + form[6])
    senior_1_1 = np.float32(senior_1_1)
    senior_1_2 = form[10] + form[7] + form[1] + form[4] + form[3] + form[2] + form[5] + form[8] + form[6]
    senior_1_2 = np.float32(senior_1_2)

    # The sum of valences or the total number of atoms having odd valences is even
    if senior_1_1 % 2 != 0 or senior_1_2 % 2 != 0:
        return False

    senior_2 = np.float32(np.sum(form))
    # The sum of valences is greater than or equal to twice the number of atoms minus 1
    if senior_1_1 < 2 * (senior_2 - 1):
        return False
    return True


@njit
def _o_p_check(form: np.array) -> bool:
    """
    check whether a formula satisfies the O/P ratio rule
    :param form: 12-dim array
    :return: True if satisfies, False otherwise
    """
    # ["C", "H", "Br", "Cl", "F", "I", "K", "N", "Na", "O", "P", "S"]
    if form[10] == 0:
        return True
    if form[9] < 3 * form[10]:
        return False
    return True


@njit
def _dbe_check(form: np.array) -> bool:
    """
    check whether a formula DBE >= 0
    :param form: 12-dim array
    :return: True if satisfies, False otherwise
    """
    dbe = form[0] + 1 - (form[1] + form[4] + form[3] + form[2] + form[5] + form[8] +
                         form[6]) / 2.0 + (form[7] + form[10]) / 2.0
    if dbe < 0:
        return False
    return True


def _adduct_loss_check(form: np.array, adduct_loss_form) -> bool:
    """
    check whether a precursor neutral formula contains the adduct loss
    :param form: 12-dim array
    :param adduct_loss_form: Formula object or None
    :return: True if contains, False otherwise
    """
    if adduct_loss_form is None:
        return True
    if np.any(form - adduct_loss_form.array < 0):
        return False
    return True


def _calc_ms1_iso_sim(cand_form, meta_feature, max_isotope_cnt) -> float:
    """
    calculate isotope similarity for a neutral precursor formula (CandidateFormula object)
    :param cand_form: CandidateFormula object
    :param meta_feature: MetaFeature object
    :param max_isotope_cnt: maximum isotope count, used for MS1 isotope pattern matching
    :return: ms1 isotope similarity
    """
    # calculate theoretical isotope pattern
    theo_isotope_pattern = calc_isotope_pattern(cand_form.charged_formula, max_isotope_cnt)

    # calculate ms1 isotope similarity
    ms1_isotope_sim = calc_isotope_similarity(meta_feature.ms1_processed.int_array, theo_isotope_pattern,
                                              max_isotope_cnt)

    return ms1_isotope_sim


def _gen_candidate_formula_from_mz(meta_feature: MetaFeature,
                                   ppm: bool, ms1_tol: float,
                                   lower_limit: np.array, upper_limit: np.array,
                                   db_mode: int, gd: dict) -> Tuple[List[CandidateFormula], List[str]]:
    """
    Generate candidate formulas for a metabolic feature with precursor mz only
    :param meta_feature: MetaFeature object
    :param ppm: whether to use ppm as the unit of tolerance
    :param ms1_tol: mz tolerance for precursor ions
    :param lower_limit: lower limit of each element
    :param upper_limit: upper limit of each element
    :param db_mode: database mode
    :param gd: global dictionary
    :return: list of candidate formulas (CandidateFormula), list of candidate formula strings
    """
    # query precursor mz
    neutral_formulas, charged_formulas = query_precursor_mass(meta_feature.mz, meta_feature.adduct,
                                                              ms1_tol, ppm, db_mode, gd)
    # filter out formulas that exceed element limits
    neutral_forms, charged_forms = [], []
    for m, f in enumerate(neutral_formulas):
        if _element_check(f.array, lower_limit, upper_limit) and _senior_rules(f.array) and _o_p_check(f.array) \
                and _dbe_check(f.array) and _adduct_loss_check(f.array, meta_feature.adduct.loss_formula):
            neutral_forms.append(f)
            charged_forms.append(charged_formulas[m])

    # convert neutral formulas into CandidateFormula objects
    cand_form_list = [CandidateFormula(formula=form, charged_formula=charged_form, exp_ms2_sum_int=0.0,
                                       db_existed=True) for form, charged_form in zip(neutral_forms, charged_forms)]
    cand_form_str_list = [form_arr_to_str(cf.formula.array) for cf in cand_form_list]

    return cand_form_list, cand_form_str_list


def _gen_candidate_formula_from_ms2(mf: MetaFeature, ppm: bool, ms1_tol: float, ms2_tol: float,
                                    lower_limit: np.array, upper_limit: np.array,
                                    db_mode: int, gd) -> Tuple[List[CandidateFormula], List[str]]:
    """
    Generate candidate formulas for a metabolic feature with MS2 data, then apply element limits
    :param mf: MetaFeature object
    :param ppm: whether to use ppm as the unit of tolerance
    :param ms1_tol: mz tolerance for precursor ions
    :param ms2_tol: mz tolerance for fragment ions / neutral losses
    :param lower_limit: lower limit of each element
    :param upper_limit: upper limit of each element
    :param db_mode: database mode
    :param gd: global dictionary
    :return: list of candidate formulas (CandidateFormula), list of candidate formula strings
    """

    # normalize MS2 intensity
    mf.ms2_processed.normalize_intensity(method='sum')

    # check whether Na and K are contained in the adduct
    na_bool = True if mf.adduct.net_formula.array[8] > 0 else False
    k_bool = True if mf.adduct.net_formula.array[6] > 0 else False

    # calculate absolute MS1 tolerance
    ms1_abs_tol = ms1_tol if not ppm else ms1_tol * mf.mz * 1e-6

    candidate_space_list = []
    existing_cand_str_list = []
    for i in range(len(mf.ms2_processed.mz_array)):
        # fragment ion m/z
        frag_mz = mf.ms2_processed.mz_array[i]
        # neutral loss m/z
        nl_mz = mf.mz - frag_mz

        # query mass in formula database
        frag_form_list, nl_form_list = _query_frag_nl_pair(frag_mz, nl_mz, mf.adduct.pos_mode, na_bool, k_bool,
                                                           ms2_tol, ppm, db_mode, gd)
        if frag_form_list is None or nl_form_list is None:
            continue

        # frag intensity
        frag_int = mf.ms2_processed.int_array[i]

        # formula stitching
        # iterate fragment formula list and neutral loss formula list
        for frag in frag_form_list:
            for nl in nl_form_list:
                # DBE check, sum of DBE should be a non-integer
                if (frag.dbe + nl.dbe) % 1 == 0 or (frag.dbe + nl.dbe) < 0 or frag.dbe < 0:
                    continue
                # sum mass check
                if abs(frag.mass + nl.mass - mf.mz) > ms1_abs_tol:
                    continue

                # generate precursor formula & check adduct M
                # NOTE: pre_form_arr is in neutral form
                pre_form_arr = (frag.array + nl.array - mf.adduct.net_formula.array) / mf.adduct.m
                valid_pre_form = _valid_precursor_array(pre_form_arr)
                if not valid_pre_form:
                    continue
                # if valid, convert to int16
                pre_form_arr = np.int16(pre_form_arr)

                # add to candidate space list
                candidate_space_list, existing_cand_str_list = _add_to_candidate_space_list(candidate_space_list,
                                                                                            existing_cand_str_list,
                                                                                            pre_form_arr.astype(int),
                                                                                            frag.array, nl.array,
                                                                                            frag_int, i)

    # element limit check, SENIOR rules, O/P check, DBE check
    candidate_list = [cs for cs in candidate_space_list
                      if _element_check(cs.pre_neutral_array, lower_limit, upper_limit)
                      and _senior_rules(cs.pre_neutral_array) and _o_p_check(cs.pre_neutral_array)
                      and _dbe_check(cs.pre_neutral_array)
                      and _adduct_loss_check(cs.pre_neutral_array, mf.adduct.loss_formula)]

    # remove candidate space variable to save memory
    del candidate_space_list

    # generate CandidateFormula object
    candidate_formula_list = [CandidateFormula(formula=Formula(cs.pre_neutral_array, 0, cs.neutral_mass),
                                               charged_formula=Formula(cs.pre_charged_array, mf.adduct.charge),
                                               exp_ms2_sum_int=cs.exp_frag_sum_int) for cs in candidate_list]
    cand_form_str_list = [form_arr_to_str(cf.formula.array) for cf in candidate_formula_list]

    return candidate_formula_list, cand_form_str_list


def _query_frag_nl_pair(frag_mz: float, nl_mz: float, pos_mode: bool, na_bool: bool, k_bool: bool,
                        ms2_tol: float, ppm: bool,
                        db_mode: int, gd) -> Tuple[Union[List[Formula], None], Union[List[Formula], None]]:
    """
    query fragment and neutral loss formulas from database
    :param frag_mz: fragment m/z
    :param nl_mz: neutral loss m/z
    :param pos_mode: positive mode or negative mode
    :param na_bool: whether Na is contained in the adduct
    :param k_bool: whether K is contained in the adduct
    :param ms2_tol: MS2 tolerance
    :param ppm: whether to use ppm as the unit of tolerance
    :param db_mode: database mode
    :param gd: global dictionary
    :return: fragment formula list, neutral loss formula list
    """
    if nl_mz < frag_mz:
        # search neutral loss first, for faster search
        nl_form_list = query_fragnl_mass(nl_mz, False, pos_mode, na_bool, k_bool,
                                         ms2_tol, ppm, db_mode, gd)
        if nl_form_list:
            frag_form_list = query_fragnl_mass(frag_mz, True, pos_mode,
                                               na_bool, k_bool, ms2_tol, ppm, db_mode, gd)
        else:
            return None, None
    else:
        frag_form_list = query_fragnl_mass(frag_mz, True, pos_mode, na_bool, k_bool,
                                           ms2_tol, ppm, db_mode, gd)
        if frag_form_list:
            nl_form_list = query_fragnl_mass(nl_mz, False, pos_mode, na_bool, k_bool,
                                             ms2_tol, ppm, db_mode, gd)
        else:
            return None, None

    return frag_form_list, nl_form_list


def _add_to_candidate_space_list(candidate_space_list: List[CandidateSpace], existing_cand_str_list: List[str],
                                 pre_form_arr: np.array, frag_arr: np.array, nl_arr: np.array,
                                 fragment_intensity: float, frag_idx: int) -> Tuple[List[CandidateSpace], List[str]]:
    """
    add a new candidate formula to the candidate space list
    :param candidate_space_list: candidate space list
    :param existing_cand_str_list: existing candidate formula string list
    :param pre_form_arr: precursor formula array
    :param frag_arr: fragment formula array
    :param nl_arr: neutral loss formula array
    :param fragment_intensity: fragment intensity
    :param frag_idx: index of the fragment in ms2_processed
    :return: updated candidate space list
    """
    # check whether the precursor formula is already in the candidate space list
    this_pre_str = form_arr_to_str(pre_form_arr)
    candidate_exist = True if this_pre_str in existing_cand_str_list else False
    # this precursor formula has not been added to the candidate space list
    if not candidate_exist:
        candidate_space_list.append(CandidateSpace(pre_form_arr, frag_arr + nl_arr,
                                                   pre_neutral_form_str=this_pre_str,
                                                   exp_frag_sum_int=fragment_intensity,
                                                   exp_frag_idx_ls=[frag_idx]))
        existing_cand_str_list.append(this_pre_str)
    else:
        # find the index of the existing candidate space
        idx = existing_cand_str_list.index(this_pre_str)
        # if the fragment is already in the exp_frag_idx_ls
        if frag_idx in candidate_space_list[idx].exp_frag_idx_ls:
            pass
        else:
            # add fragment intensity
            candidate_space_list[idx].exp_frag_sum_int += fragment_intensity
            # add fragment idx
            candidate_space_list[idx].exp_frag_idx_ls.append(frag_idx)

    return candidate_space_list, existing_cand_str_list


@njit
def _valid_precursor_array(pre_arr: np.array) -> bool:
    """
    check adduct M
    :param pre_arr: precursor formula array
    :return: precursor formula array or None
    """
    for i in range(len(pre_arr)):
        if pre_arr[i] < 0 or pre_arr[i] % 1 != 0:
            return False
    return True


def _merge_cand_form_list(ms1_cand_list: List[CandidateFormula], ms2_cand_list: List[CandidateFormula],
                          ms1_cand_str_list: List[str], ms2_cand_str_list: List[str]) -> List[CandidateFormula]:
    """
    Merge MS1 and MS2 candidate formula lists.
    Map MS2 candidate formulas to MS1 candidate formulas (db_existed=True)
    :param ms1_cand_list: candidate formula list from MS1 mz search
    :param ms2_cand_list: candidate formula list from MS2 interrogation
    :param ms1_cand_str_list: candidate formula string list from MS1 mz search
    :param ms2_cand_str_list: candidate formula string list from MS2 interrogation
    :return: merged candidate formula list, remove duplicates
    """
    out_list = ms1_cand_list.copy()
    for m, cf in enumerate(ms2_cand_list):
        found = False
        for n, cf2 in enumerate(ms1_cand_list):
            if ms1_cand_str_list[n] == ms2_cand_str_list[m]:
                found = True
                break
        if not found:
            out_list.append(cf)

    return out_list


def _fill_in_db_existence(ms1_cand_list: List[CandidateFormula], ms2_cand_list: List[CandidateFormula],
                          ms1_cand_str_list: List[str], ms2_cand_str_list: List[str]) -> List[CandidateFormula]:
    """
    Fill in DB existence for MS2 candidate formulas.
    :param ms1_cand_list: candidate formula list from MS1 mz search
    :param ms2_cand_list: candidate formula list from MS2 interrogation
    :param ms1_cand_str_list: candidate formula string list from MS1 mz search
    :param ms2_cand_str_list: candidate formula string list from MS2 interrogation
    :return: ms2 candidate formula list with db_existed filled in
    """
    for m, cf in enumerate(ms2_cand_list):
        found = False
        for n, cf2 in enumerate(ms1_cand_list):
            if ms1_cand_str_list[n] == ms2_cand_str_list[m]:
                found = True
                break
        if found:
            ms2_cand_list[m].db_existed = True

    return ms2_cand_list


def _retain_top_cand_form(cf_list: List[CandidateFormula], ms1_tol: float, ms1_iso_available: bool,
                          ms2_available: bool, top_n: int) -> List[CandidateFormula]:
    """
    Retain top candidate formulas by explained MS2 intensity, then by mz difference.
    :param cf_list: candidate formula list
    :param ms1_tol: MS1 tolerance
    :param ms1_iso_available: whether MS1 isotope data is available
    :param ms2_available: whether MS2 data is available
    :param top_n: number of top candidate formulas to retain
    :return: retained candidate formula list
    """
    if len(cf_list) <= top_n:
        return cf_list
    else:
        if ms1_iso_available:
            # sort candidate list by logp mz difference * 0.05 + ms1_iso_sim
            cf_list = sorted(cf_list, key=lambda x: (_calc_log_p_norm(x.mz_error, ms1_tol / 3) * 0.05 +
                                                     x.ms1_isotope_similarity), reverse=True)
        else:
            if ms2_available:
                # sort candidate list by mz_error_log_p + 4 * exp_frag_int_pnt
                cf_list = sorted(cf_list, key=lambda x: (_calc_log_p_norm(x.mz_error, ms1_tol / 3) + 4 * x.exp_ms2_sum_int),
                                 reverse=True)
            else:
                # sort candidate list by abs mz_error, ascending
                cf_list = sorted(cf_list, key=lambda x: abs(x.mz_error), reverse=False)

        return cf_list[:top_n]


@njit
def _form_array_equal(arr1: np.array, arr2: np.array) -> bool:
    """
    check whether two formula arrays are equal
    :param arr1: 12-dim array
    :param arr2: 12-dim array
    :return: True if equal, False otherwise
    """
    return True if np.equal(arr1, arr2).all() else False


def assign_subformula_cand_form(mf: MetaFeature, ppm: bool, ms2_tol: float) -> MetaFeature:
    """
    Assign subformula to all candidate formulas in a MetaFeature object.
    :param mf: MetaFeature object
    :param ppm: whether to use ppm as the unit of tolerance
    :param ms2_tol: mz tolerance for fragment ions / neutral losses
    :return: MetaFeature object
    """

    for k, cf in enumerate(mf.candidate_formula_list):
        # enumerate all subformulas
        subform_arr = enumerate_subformula(cf.charged_formula.array)

        # mono mass
        mass_arr = _calc_subform_mass(subform_arr, mf.adduct.charge)
        # assign ms2 explanation
        mf.candidate_formula_list[k] = _assign_ms2_explanation(mf, cf, cf.charged_formula.array, subform_arr, mass_arr,
                                                               ppm, ms2_tol)

    return mf


@njit
def _calc_subform_mass(subform_arr: np.array, adduct_charge: int) -> np.array:
    """
    Calculate mass of each subformula.
    :param subform_arr: 2D array, each row is a subformula array
    :param adduct_charge: adduct charge
    :return: 1D array, mass of each subformula
    """
    mass_arr = np.empty(subform_arr.shape[0], dtype=np.float32)
    ele_mass_arr = np.array([12.000000, 1.007825, 78.918336, 34.968853, 18.998403, 126.904473, 38.963707, 14.003074,
                             22.989769, 15.994915, 30.973762, 31.972071], dtype=np.float32)
    for i in range(subform_arr.shape[0]):
        # element wise multiplication
        mass_arr[i] = np.sum(subform_arr[i, :] * ele_mass_arr, dtype=np.float32) - np.float32(adduct_charge * 0.0005486)
    return mass_arr


@njit
def _dbe_subform_filter(subform_arr: np.array, cutoff: float) -> np.array:
    """
    Filter subformulas by DBE.
    :param subform_arr: 2D array, each row is a subformula array
    :return: boolean array
    """
    dbe_arr = subform_arr[:, 0] + 1 - (subform_arr[:, 1] + subform_arr[:, 4] + subform_arr[:, 3] + subform_arr[:, 2]
                                       + subform_arr[:, 5] + subform_arr[:, 8] + subform_arr[:, 6]) / 2 + \
              (subform_arr[:, 7] + subform_arr[:, 10]) / 2
    dbe_bool_arr = dbe_arr >= cutoff
    return dbe_bool_arr


@njit
def _senior_subform_filter(subform_arr: np.array) -> np.array:
    """
    Filter subformulas by SENIOR rules.
    :param subform_arr: 2D array, each row is a subformula array
    :return: boolean array
    """
    senior_1_1_arr = 6 * subform_arr[:, 11] + 5 * subform_arr[:, 10] + 4 * subform_arr[:, 0] + \
                     3 * subform_arr[:, 7] + 2 * subform_arr[:, 9] + subform_arr[:, 1] + subform_arr[:, 4] + \
                     subform_arr[:, 3] + subform_arr[:, 2] + subform_arr[:, 5] + subform_arr[:, 8] + subform_arr[:, 6]
    senior_2_arr = np.sum(subform_arr, axis=1)
    senior_bool_arr = (senior_1_1_arr >= 2 * (senior_2_arr - 1))

    return senior_bool_arr


@njit
def _valid_subform_check(subform_arr: np.array, pre_charged_arr: np.array) -> np.array:
    """
    Check whether a subformula (frag and loss) is valid. e.g., 'C2', 'N4', 'P2'; O >= 2*P
    :param subform_arr: 2D array, each row is a subformula array
    :return: boolean array
    """
    # for frag or loss
    frag_atom_sum = np.sum(subform_arr, axis=1)
    loss_form_arr = pre_charged_arr - subform_arr
    loss_atom_sum = np.sum(loss_form_arr, axis=1)
    invalid_bool_arr = (frag_atom_sum == subform_arr[:, 0]) | (frag_atom_sum == subform_arr[:, 7]) | \
                       (frag_atom_sum == subform_arr[:, 10]) | (loss_atom_sum == loss_form_arr[:, 0]) | \
                       (loss_atom_sum == loss_form_arr[:, 7]) | (loss_atom_sum == loss_form_arr[:, 10])

    # O >= 2*P if P > 0
    invalid_o_p_frag_bool_arr = (subform_arr[:, 9] < 2 * subform_arr[:, 10]) & (subform_arr[:, 10] > 0)
    invalid_o_p_loss_bool_arr = (loss_form_arr[:, 9] < 2 * loss_form_arr[:, 10]) & (loss_form_arr[:, 10] > 0)

    invalid_bool_arr = invalid_bool_arr | invalid_o_p_frag_bool_arr | invalid_o_p_loss_bool_arr

    return ~invalid_bool_arr


def _assign_ms2_explanation(mf: MetaFeature, cf: CandidateFormula, pre_charged_arr: np.array,
                            subform_arr: np.array, mass_arr: np.array,
                            ppm: bool, ms2_tol: float) -> CandidateFormula:
    """
    Assign MS2 explanation to a candidate formula.
    :param mf: MetaFeature object
    :param cf: CandidateFormula object
    :param pre_charged_arr: precursor charged array
    :param subform_arr: 2D array, each row is a subformula array
    :param mass_arr: 1D array, mass of each subformula
    :param ppm: whether to use ppm as the unit of tolerance
    :param ms2_tol: mz tolerance for fragment ions / neutral losses
    :return: CandidateFormula object
    """
    candidate_space = None
    ion_mode_int = 1 if mf.adduct.pos_mode else -1
    for i in range(len(mf.ms2_processed.mz_array)):
        # retrieve all indices of mass within tolerance
        this_ms2_tol = ms2_tol if not ppm else ms2_tol * mf.ms2_processed.mz_array[i] * 1e-6
        idx_list = np.where(abs(mf.ms2_processed.mz_array[i] - mass_arr) <= this_ms2_tol)[0]

        if len(idx_list) == 0:
            continue

        # retrieve all subformulas within tolerance
        this_subform_arr = subform_arr[idx_list, :]
        this_mass = mass_arr[idx_list]

        # dbe filter (DBE >= 0)
        bool_arr_1 = _dbe_subform_filter(this_subform_arr, 0)

        # SENIOR rules filter, a soft version
        bool_arr_2 = _senior_subform_filter(this_subform_arr)

        # valid subformula check
        bool_arr_3 = _valid_subform_check(this_subform_arr, pre_charged_arr)

        # combine filters
        bool_arr = bool_arr_1 & bool_arr_2 & bool_arr_3
        this_subform_arr = this_subform_arr[bool_arr, :]
        this_mass = this_mass[bool_arr]

        # if no valid subformula, skip
        if this_subform_arr.shape[0] == 0:
            continue

        frag_exp = FragExplanation(mf.ms2_processed.idx_array[i],
                                   Formula(this_subform_arr[0, :], ion_mode_int, this_mass[0]),
                                   Formula(pre_charged_arr - this_subform_arr[0, :], 0))
        # add all subformulas
        if len(this_mass) > 1:
            for j in range(1, len(this_mass)):
                frag_exp.add_frag_nl(Formula(this_subform_arr[j, :], ion_mode_int, this_mass[j]),
                                     Formula(pre_charged_arr - this_subform_arr[j, :], 0))

        if candidate_space is None:
            # create CandidateSpace object
            candidate_space = CandidateSpace(cf.formula.array, pre_charged_arr, frag_exp_ls=[frag_exp])
        else:
            candidate_space.add_frag_exp(frag_exp)

    # if no MS2 explanation, return original candidate formula
    if not candidate_space:
        return cf

    # refine MS2 explanation
    ms2_iso_tol = ms2_tol if not ppm else ms2_tol * mf.mz * 1e-6
    ms2_iso_tol = max(ms2_iso_tol, 0.02)
    candidate_form = candidate_space.refine_explanation(mf, ms2_iso_tol)

    # copy other attributes
    candidate_form.ms1_isotope_similarity = cf.ms1_isotope_similarity
    candidate_form.formula_feature_array = cf.formula_feature_array

    return candidate_form


def assign_subformula(ms2_mz: List, precursor_formula: str, adduct: str,
                      ms2_tol: float = 10, ppm: bool = True,
                      dbe_cutoff: float = 0.0) -> Union[List[SubformulaResult], None]:
    """
    Assign subformulas to a given MS2 spectrum with a given precursor formula and adduct. Radical ions are considered.
    :param ms2_mz: MS2 m/z list
    :param precursor_formula: precursor formula string, uncharged
    :param adduct: adduct string, e.g., [M+H]+
    :param ms2_tol: MS2 tolerance
    :param ppm: whether MS2 tolerance is in ppm
    :param dbe_cutoff: float, DBE cutoff
    :return: a list of SubformulaResult objects
    """
    assert len(ms2_mz) > 0, "MS2 m/z list is empty."

    # convert precursor formula to array
    form_arr = read_formula(precursor_formula)
    if form_arr is None:
        return None

    # convert adduct to array
    valid_adduct, pos_mode = check_adduct(adduct)
    if not valid_adduct:
        raise ValueError("Invalid adduct string.")
    ion = Adduct(adduct, pos_mode)
    ion_mode_int = 1 if pos_mode else -1

    # enumerate all subformulas
    pre_charged_arr = form_arr * ion.m + ion.net_formula.array
    subform_arr = enumerate_subformula(pre_charged_arr)

    # mono mass
    mass_arr = _calc_subform_mass(subform_arr, ion.charge)

    # assign ms2 explanation
    out_list = []
    for k, mz in enumerate(ms2_mz):
        # retrieve all indices of mass within tolerance
        this_ms2_tol = ms2_tol if not ppm else ms2_tol * mz * 1e-6
        idx_list = np.where(abs(mz - mass_arr) <= this_ms2_tol)[0]

        if len(idx_list) == 0:
            out_list.append(SubformulaResult(k, []))
            continue

        # retrieve all subformulas within tolerance
        this_subform_arr = subform_arr[idx_list, :]
        this_mass = mass_arr[idx_list]

        # dbe filter (DBE >= dbe_cutoff)
        bool_arr_1 = _dbe_subform_filter(this_subform_arr, dbe_cutoff)

        # SENIOR rules filter, a soft version
        bool_arr_2 = _senior_subform_filter(this_subform_arr)

        # valid subformula check
        bool_arr_3 = _valid_subform_check(this_subform_arr, pre_charged_arr)

        # combine filters
        bool_arr = bool_arr_1 & bool_arr_2 & bool_arr_3
        this_subform_arr = this_subform_arr[bool_arr, :]
        this_mass = this_mass[bool_arr]

        # if no valid subformula, skip
        if this_subform_arr.shape[0] == 0:
            out_list.append(SubformulaResult(k, []))
            continue

        # create SubformulaResult object
        subform_list = []
        for j in range(this_subform_arr.shape[0]):
            form = Formula(this_subform_arr[j, :], ion_mode_int, this_mass[j])
            subform_list.append(FormulaResult(form.__str__(), form.mass, mz))
        out_list.append(SubformulaResult(k, subform_list))

    return out_list



# ==============================================================================
# Copyright (C) 2024 Shipei Xing <philipxsp@hotmail.com>
#
# Licensed under the Apache License 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at https://github.com/Philipbear/msbuddy/blob/main/LICENSE
# ==============================================================================
"""
File: load.py
Author: Shipei Xing
Email: philipxsp@hotmail.com
GitHub: Philipbear
Description: load databases and data files
"""

import logging
from json import loads as loads
from pathlib import Path
from typing import List, Union

import numpy as np
from gdown import download as download
from joblib import load as j_load
from requests import get

from msbuddy.base import MetaFeature, Spectrum

logging.basicConfig(level=logging.INFO)

current_db_version = 'v0.2.4'
current_model_version = 'v0.3.0'


def check_download_joblibload(url: str, path):
    """
    check if the file exists, if not, download from url, and load
    :param url: url to download
    :param path: path to save
    :return: loaded object
    """
    if not path.exists() or path.stat().st_size < 10 ** 3:
        download(url, str(path))

    return j_load(path)


def init_db() -> dict:
    """
    init databases used in the project
    :return: global_dict
    """
    # get root path
    root_path = Path(__file__).parent

    # create data folder if not exists
    data_path = root_path / 'data'
    data_path.mkdir(parents=True, exist_ok=True)

    global_dict = dict()

    # load common_loss_db, common_frag_db
    db_name = 'common_db_' + current_db_version + '.joblib'
    global_dict['common_loss_db'], global_dict['common_frag_db'] = (
        check_download_joblibload(
            'https://github.com/Philipbear/msbuddy/releases/download/msbuddy_data_v0.2.4/common_db_v0.2.4.joblib',
            data_path / db_name))

    # formula_db
    db_name = 'formula_db_' + current_db_version + '.joblib'
    basic_db, halogen_db = (
        check_download_joblibload(
            'https://github.com/Philipbear/msbuddy/releases/download/msbuddy_data_v0.2.4/formula_db_v0.2.4.joblib',
            data_path / db_name))

    global_dict['basic_db_mass'], global_dict['basic_db_formula'], global_dict['basic_db_idx'] = basic_db
    global_dict['halogen_db_mass'], global_dict['halogen_db_formula'], global_dict['halogen_db_idx'] = halogen_db

    return global_dict


def init_ml_models(global_dict) -> dict:
    """
    init ml models if not exists
    :return: global_dict
    """

    # check if ml models exist
    if 'model_ms1_ms2' in global_dict:
        return global_dict

    # get root path
    root_path = Path(__file__).parent

    # create data folder if not exists
    data_path = root_path / 'data'
    data_path.mkdir(parents=True, exist_ok=True)

    # load ml
    ml_name = 'ml_' + current_model_version + '.joblib'
    (global_dict['model_ms1_ms2'], global_dict['model_noms1_ms2'], global_dict['model_ms1_noms2'], global_dict[
        'model_noms1_noms2'], global_dict['platt_a_0'], global_dict['platt_b_0'], global_dict['platt_a_1'],
     global_dict['platt_b_1'], global_dict['platt_a_2'], global_dict['platt_b_2'], global_dict['platt_a_3'],
     global_dict['platt_b_3']) = (
        check_download_joblibload(
            'https://github.com/Philipbear/msbuddy/releases/download/msbuddy_data_v0.3.0/ml_v0.3.0.joblib',
            data_path / ml_name))

    return global_dict


def load_mgf(file_path) -> List[MetaFeature]:
    """
    read mgf file
    :param file_path: path to mgf file
    :return: list of MetaFeature
    """
    with open(file_path, 'r') as file:
        # create meta_feature_list
        meta_feature_list = []
        cnt = 0
        for line in file:
            # empty line
            _line = line.strip()  # remove leading and trailing whitespace
            if not _line:
                continue
            elif line.startswith('BEGIN IONS'):
                # initialize a new spectrum entry
                mz_arr = np.array([])
                int_arr = np.array([])
                precursor_mz = None
                identifier = None
                charge = None
                pos_mode = None
                ms2_spec = True
                rt = None
                adduct_str = None
            elif line.startswith('END IONS'):
                # create a new MetaFeature
                if precursor_mz is None:
                    raise ValueError('No precursor mz found.')
                if identifier is None:
                    identifier = cnt
                if charge is None:
                    charge = 1 if pos_mode else -1
                elif charge == 0:
                    charge = 1 if pos_mode else -1
                else:
                    charge = abs(charge) if pos_mode else -abs(charge)

                # # if no peaks found, skip
                # if mz_arr.size == 0:
                #     continue

                # create MetaFeature object if the same identifier does not exist
                mf_idx = None
                for idx, mf in enumerate(meta_feature_list):
                    if mf.identifier == identifier:
                        mf_idx = idx
                        break

                # if the same identifier exists, add to the existing MetaFeature
                if mf_idx is not None:
                    if ms2_spec and meta_feature_list[mf_idx].ms2_raw is None:
                        meta_feature_list[mf_idx].ms2_raw = Spectrum(mz_arr, int_arr) if mz_arr.size > 0 else None
                    elif ms2_spec is False and meta_feature_list[mf_idx].ms1_raw is None:
                        meta_feature_list[mf_idx].ms1_raw = Spectrum(mz_arr, int_arr) if mz_arr.size > 0 else None
                    continue
                # if the same identifier does not exist, create a new MetaFeature
                else:
                    if ms2_spec:
                        mf = MetaFeature(mz=precursor_mz,
                                         charge=charge,
                                         rt=rt,
                                         adduct=adduct_str,
                                         ms2=Spectrum(mz_arr, int_arr) if mz_arr.size > 0 else None,
                                         identifier=identifier)
                    else:
                        mf = MetaFeature(mz=precursor_mz,
                                         charge=charge,
                                         rt=rt,
                                         adduct=adduct_str,
                                         ms1=Spectrum(mz_arr, int_arr) if mz_arr.size > 0 else None,
                                         identifier=identifier)
                    meta_feature_list.append(mf)
                    cnt += 1
                continue
            else:
                # if line contains '=', it is a key-value pair
                if '=' in _line:
                    # split by first '=', in case of multiple '=' in the line
                    key, value = _line.split('=', 1)
                    key, value = key.strip(), value.strip()
                    # if key (into all upper case) is 'PEPMASS', it is precursor mz
                    if key.upper() in ['PEPMASS', 'PRECURSOR_MZ', 'PRECURSORMZ']:
                        precursor_mz = float(value)
                    # identifier
                    elif key.upper() in ['TITLE', 'FEATURE_ID', 'SPECTRUMID', 'SPECTRUM_ID']:
                        identifier = value.strip()
                    # if key is 'CHARGE' and charge is not set, it is charge
                    elif key.upper() == 'CHARGE':
                        if '-' in value:
                            pos_mode = False
                            value = value.replace('-', '')
                            charge = -int(value)
                        else:
                            pos_mode = True
                            value = value.replace('+', '')
                            charge = int(value)
                    # if key is 'ION', it is adduct type
                    elif key.upper() in ['ION', 'IONTYPE', 'ION_TYPE', 'ADDUCT', 'ADDUCTTYPE', 'ADDUCT_TYPE']:
                        adduct_str = value
                    # if key is 'IONMODE', it is ion mode
                    elif key.upper() in ['IONMODE', 'ION_MODE']:
                        if value.upper() in ['POSITIVE', 'POS', 'P']:
                            pos_mode = True
                        elif value.upper() in ['NEGATIVE', 'NEG', 'N']:
                            pos_mode = False
                    # if key is 'MSLEVEL', it is ms level
                    elif key.upper() == 'MSLEVEL':
                        if value == '1':
                            ms2_spec = False
                    # if key is 'RTINSECONDS', it is rt
                    elif key.upper() == 'RTINSECONDS' and value != '':
                        rt = float(value)
                    elif key.upper() == 'RTINMINUTES' and value != '':
                        rt = float(value) * 60
                else:
                    # if no '=', it is a spectrum pair, split by '\t' or ' '
                    this_mz, this_int = _line.split()
                    mz_arr = np.append(mz_arr, float(this_mz))
                    int_arr = np.append(int_arr, float(this_int))

    return meta_feature_list


def _load_usi(usi: str, adduct: Union[str, None] = None) -> MetaFeature:
    """
    Read from a USI string and return a MetaFeature object.
    The GNPS API is used to get the spectrum from the USI.
    Citation: Universal MS/MS Visualization and Retrieval with the Metabolomics Spectrum Resolver Web Service.
    Wout Bittremieux et al. doi: 10.1101/2020.05.09.085000
    :param usi: USI string
    :param adduct: adduct string
    :return: MetaFeature object
    """
    # get spectrum from USI
    url = 'https://api.metabolomics-usi.gnps2.org/json/?usi1=' + usi
    response = get(url, timeout=10)
    json_data = loads(response.text)

    # check if the USI is valid
    if 'error' in json_data:
        raise ValueError

    # get adduct
    if adduct == '':
        adduct = None

    # valid: dict_keys(['n_peaks', 'peaks', 'precursor_charge', 'precursor_mz', 'splash'])
    # ion mode
    charge = json_data['precursor_charge']
    if charge == 0 and adduct is not None:
        pos_mode = str(adduct)[-1] != '-'  # use adduct if charge is 0
        charge = 1 if pos_mode else -1

    ms2_mz = np.array(json_data['peaks'])[:, 0]
    ms2_int = np.array(json_data['peaks'])[:, 1]

    data = MetaFeature(mz=json_data['precursor_mz'],
                       charge=charge,
                       adduct=adduct,
                       ms2=Spectrum(ms2_mz, ms2_int),
                       identifier=usi)
    return data


def load_usi(usi_list: Union[str, List[str]],
             adduct_list: Union[None, str, List[str]] = None) -> List[MetaFeature]:
    """
    Read from a sequence of USI strings and return a list of MetaFeature objects.
    Invalid USI strings are discarded.
    The GNPS API is used to get the spectrum from the USI.
    Citation: Universal MS/MS Visualization and Retrieval with the Metabolomics Spectrum Resolver Web Service.
    Wout Bittremieux et al. doi: 10.1101/2020.05.09.086066.
    See https://ccms-ucsd.github.io/GNPSDocumentation/api/#experimental-or-library-spectrum-by-usi for details.
     ---------------------------------------------------------
    :param usi_list: List of USI string or a single USI string
    :param adduct_list: adduct string, e.g. [M+H]+
    :return: List of MetaFeature objects
    """

    data_list = []

    # if usi_list is a single string, convert it to a list
    if isinstance(usi_list, str):
        usi_list = [usi_list]
        if adduct_list is not None:
            adduct_list = [adduct_str.strip() for adduct_str in adduct_list]

    if adduct_list is None:
        adduct_list = [None] * len(usi_list)
    elif len(adduct_list) != len(usi_list):
        logging.warning('adduct_list and usi_list must have the same length. Default adducts are used.')

    usi_list = [usi.strip() for usi in usi_list]

    # retrieve indices of unique USIs from the list
    seen = {}
    unique_indices = []
    for idx, item in enumerate(usi_list):
        if item not in seen:
            seen[item] = True
            unique_indices.append(idx)
        else:
            logging.warning('Duplicate USI: ' + item + '. Only the first occurrence is used.')
    usi_list_unique = [usi_list[idx] for idx in unique_indices]
    adduct_list_unique = [adduct_list[idx] for idx in unique_indices]

    # load data
    for usi, adduct in zip(usi_list_unique, adduct_list_unique):
        try:
            data_list.append(_load_usi(usi, adduct))
        except:
            logging.warning('Invalid USI: ' + usi)
            continue
    return data_list


if __name__ == '__main__':
    init_db()
    # compile all these databases
    # import joblib
    # basic_db_mass = j_load('../db_prep/basic_db_mass.joblib')
    # basic_db_formula = j_load('../db_prep/basic_db_formula.joblib')
    # basic_db_idx = j_load('../db_prep/basic_db_idx.joblib')
    # halogen_db_mass = j_load('../db_prep/halogen_db_mass.joblib')
    # halogen_db_formula = j_load('../db_prep/halogen_db_formula.joblib')
    # halogen_db_idx = j_load('../db_prep/halogen_db_idx.joblib')
    #
    # basic_db = [basic_db_mass, basic_db_formula, basic_db_idx]
    # halogen_db = [halogen_db_mass, halogen_db_formula, halogen_db_idx]
    # formula_db = [basic_db, halogen_db]
    #
    # joblib.dump(formula_db, "data/formula_db.joblib")

    # common_db
    # common_loss_db = j_load('data/common_loss.joblib')
    # common_frag_db = j_load('data/common_frag.joblib')
    #
    # common_db = [common_loss_db, common_frag_db]
    # joblib.dump(common_db, "data/common_db.joblib")

    # # model
    # model_ms1_ms2 = j_load('data/model_ms1_ms2.joblib')
    # model_noms1_ms2 = j_load('data/model_ms2.joblib')
    # model_ms1_noms2 = j_load('data/model_ms1.joblib')
    # model_noms1_noms2 = j_load('data/model.joblib')
    #
    # ml = [model_ms1_ms2, model_noms1_ms2, model_ms1_noms2, model_noms1_noms2, 1.898745, -2.012396, 1.930475,
    # -2.617705, 1.428206, -1.237494, 1.610414, -2.045858]
    # joblib.dump(ml, "data/ml_v0.3.0.joblib")



# ==============================================================================
# Copyright (C) 2024 Shipei Xing <philipxsp@hotmail.com>
#
# Licensed under the Apache License 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at https://github.com/Philipbear/msbuddy/blob/main/LICENSE
# ==============================================================================
"""
File: main_cmd.py
Author: Shipei Xing
Email: philipxsp@hotmail.com
GitHub: Philipbear
Description: Command line interface for msbuddy.
"""

import argparse
import pathlib

import pandas as pd

from msbuddy.main import Msbuddy, MsbuddyConfig


def main():
    parser = argparse.ArgumentParser(description="msbuddy command line interface (version 0.3.7)")
    parser.add_argument('-mgf', type=str, help='Path to the MGF file.')
    parser.add_argument('-usi', type=str, help='A single USI string.')
    parser.add_argument('-csv', type=str,
                        help='Path to the CSV file containing USI strings in the first column (no header row).')
    parser.add_argument('-output', '-o', type=str, help='The output file path.')
    parser.add_argument('-details', '-d', action='store_true',
                        help='Store true. Whether to write detailed results. Default: detailed results are not written.')
    parser.add_argument('-ms_instr', '-ms', type=str, default=None,
                        help='MS instrument type. Supported types: orbitrap, qtof, fticr.')
    parser.add_argument('-use_Da', action='store_true',
                        help='Store true. Whether to use Dalton for mass tolerance. Default: ppm is used.')
    parser.add_argument('-ms1_tol', type=float, default=5, help='MS1 tolerance. Default: 5.')
    parser.add_argument('-ms2_tol', type=float, default=10, help='MS2 tolerance. Default: 10.')
    parser.add_argument('-halogen', '-hal', action='store_true',
                        help='Store true. Whether to consider halogen atoms FClBrI. Default: halogen atoms are not considered.')
    parser.add_argument('-parallel', '-p', action='store_true',
                        help='Store true. Whether to use parallel computing. Default: parallel computing is disabled.')
    parser.add_argument('-n_cpu', type=int, default=-1, help='Number of CPUs to use. Default: -1, use all CPUs.')
    parser.add_argument('-timeout_secs', '-t', type=int, default=300, help='Timeout in seconds. Default: 300.')
    parser.add_argument('-batch_size', '-bs', type=int, default=5000,
                        help='Batch size. Default: 5000. A larger batch size needs more memory, but is faster.')
    parser.add_argument('-c_min', type=int, default=0, help='Minimum number of C atoms. Default: 0.')
    parser.add_argument('-c_max', type=int, default=80, help='Maximum number of C atoms. Default: 80.')
    parser.add_argument('-h_min', type=int, default=0, help='Minimum number of H atoms. Default: 0.')
    parser.add_argument('-h_max', type=int, default=150, help='Maximum number of H atoms. Default: 150.')
    parser.add_argument('-n_min', type=int, default=0, help='Minimum number of N atoms. Default: 0.')
    parser.add_argument('-n_max', type=int, default=20, help='Maximum number of N atoms. Default: 20.')
    parser.add_argument('-o_min', type=int, default=0, help='Minimum number of O atoms. Default: 0.')
    parser.add_argument('-o_max', type=int, default=30, help='Maximum number of O atoms. Default: 30.')
    parser.add_argument('-p_min', type=int, default=0, help='Minimum number of P atoms. Default: 0.')
    parser.add_argument('-p_max', type=int, default=10, help='Maximum number of P atoms. Default: 10.')
    parser.add_argument('-s_min', type=int, default=0, help='Minimum number of S atoms. Default: 0.')
    parser.add_argument('-s_max', type=int, default=15, help='Maximum number of S atoms. Default: 15.')
    parser.add_argument('-f_min', type=int, default=0, help='Minimum number of F atoms. Default: 0.')
    parser.add_argument('-f_max', type=int, default=20, help='Maximum number of F atoms. Default: 20.')
    parser.add_argument('-cl_min', type=int, default=0, help='Minimum number of Cl atoms. Default: 0.')
    parser.add_argument('-cl_max', type=int, default=15, help='Maximum number of Cl atoms. Default: 15.')
    parser.add_argument('-br_min', type=int, default=0, help='Minimum number of Br atoms. Default: 0.')
    parser.add_argument('-br_max', type=int, default=10, help='Maximum number of Br atoms. Default: 10.')
    parser.add_argument('-i_min', type=int, default=0, help='Minimum number of I atoms. Default: 0.')
    parser.add_argument('-i_max', type=int, default=10, help='Maximum number of I atoms. Default: 10.')
    parser.add_argument('-isotope_bin_mztol', type=float, default=0.02,
                        help='m/z tolerance for isotope binning, used for MS1 isotope pattern, in Dalton. Default: 0.02.')
    parser.add_argument('-max_isotope_cnt', type=int, default=4,
                        help='Maximum isotope count, used for MS1 isotope pattern. Default: 4.')
    parser.add_argument('-rel_int_denoise_cutoff', type=float, default=0.01,
                        help='Relative intensity cutoff, used for MS2 denoise. Default: 0.01.')
    parser.add_argument('-top_n_per_50_da', type=int, default=6,
                        help='Top n peaks per 50 Da, used for MS2 denoise. Default: 6.')

    args = parser.parse_args()

    # run msbuddy
    # create a MsbuddyConfig object
    msb_config = MsbuddyConfig(
        ms_instr=args.ms_instr,
        ppm=~args.use_Da,
        ms1_tol=args.ms1_tol, ms2_tol=args.ms2_tol, halogen=args.halogen,
        parallel=args.parallel, n_cpu=args.n_cpu,
        timeout_secs=args.timeout_secs, batch_size=args.batch_size,
        c_range=(args.c_min, args.c_max), h_range=(args.h_min, args.h_max), n_range=(args.n_min, args.n_max),
        o_range=(args.o_min, args.o_max), p_range=(args.p_min, args.p_max), s_range=(args.s_min, args.s_max),
        f_range=(args.f_min, args.f_max), cl_range=(args.cl_min, args.cl_max), br_range=(args.br_min, args.br_max),
        i_range=(args.i_min, args.i_max),
        isotope_bin_mztol=args.isotope_bin_mztol, max_isotope_cnt=args.max_isotope_cnt,
        rel_int_denoise_cutoff=args.rel_int_denoise_cutoff, top_n_per_50_da=args.top_n_per_50_da
    )

    if args.output:
        output_path = pathlib.Path(args.output)
    elif args.mgf or args.csv:
        # use the parent directory of the input file as the output directory
        output_path = pathlib.Path(args.mgf if args.mgf else args.csv).parent / 'msbuddy_output'
    else:
        raise ValueError('Please specify the output path.')

    engine = Msbuddy(msb_config)

    if args.mgf:
        engine.load_mgf(args.mgf)
    elif args.usi:
        engine.load_usi([args.usi])
    elif args.csv:
        # read and load CSV file, treat empty cells as empty strings
        df = pd.read_csv(args.csv, keep_default_na=False, na_values=None)
        # if df has >1 columns, treat the 2nd column as adduct strings
        if df.shape[1] > 1:
            engine.load_usi(usi_list=df.iloc[:, 0].tolist(),
                            adduct_list=df.iloc[:, 1].tolist())
        else:
            engine.load_usi(df.iloc[:, 0].tolist())
    else:
        raise ValueError('Please specify the input data source.')

    engine.annotate_formula_cmd(output_path, write_details=args.details)

    print('Job finished.')


if __name__ == '__main__':
    main()



# ==============================================================================
# Copyright (C) 2024 Shipei Xing <s1xing@health.ucsd.edu>
#
# Licensed under the Apache License 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at https://github.com/Philipbear/msbuddy/blob/main/LICENSE
# ==============================================================================
"""
File: main.py
Author: Shipei Xing
Email: s1xing@health.ucsd.edu
GitHub: Philipbear
Description: main class for msbuddy
"""

import logging
import pathlib
import sys
from multiprocessing import Pool, cpu_count
from typing import Tuple, Union, List

import numpy as np
import pandas as pd
from tqdm import tqdm

from msbuddy.base import MetaFeature, Adduct, check_adduct, senior_rules
from msbuddy.cand import gen_candidate_formula, assign_subformula_cand_form
from msbuddy.export import write_batch_results_cmd
from msbuddy.load import init_db, init_ml_models, load_usi, load_mgf
from msbuddy.ml import predict_formula_probability, calc_fdr
from msbuddy.query import query_neutral_mass, query_precursor_mass
from msbuddy.utils import form_arr_to_str, FormulaResult

logging.basicConfig(level=logging.WARNING)

# global variable containing shared data
global shared_data_dict


class MsbuddyConfig:
    """
    msbuddy configuration class
    """

    def __init__(self,
                 ms_instr: str = None,
                 ppm: bool = True,
                 ms1_tol: float = 5,
                 ms2_tol: float = 10,
                 halogen: bool = False,
                 parallel: bool = False,
                 n_cpu: int = -1,
                 timeout_secs: float = 300,
                 batch_size: int = 1000,
                 c_range: Tuple[int, int] = (0, 80),
                 h_range: Tuple[int, int] = (0, 150),
                 n_range: Tuple[int, int] = (0, 20),
                 o_range: Tuple[int, int] = (0, 30),
                 p_range: Tuple[int, int] = (0, 10),
                 s_range: Tuple[int, int] = (0, 15),
                 f_range: Tuple[int, int] = (0, 20),
                 cl_range: Tuple[int, int] = (0, 15),
                 br_range: Tuple[int, int] = (0, 10),
                 i_range: Tuple[int, int] = (0, 10),
                 isotope_bin_mztol: float = 0.02, max_isotope_cnt: int = 4,
                 rel_int_denoise_cutoff: float = 0.01,
                 top_n_per_50_da: int = 6):
        """
        :param ms_instr: mass spectrometry instrument, one of "orbitrap, "fticr", "qtof".
        :param ppm: whether ppm is used for m/z tolerance
        :param ms1_tol: MS1 m/z tolerance
        :param ms2_tol: MS2 m/z tolerance
        :param halogen: whether to include halogen atoms; if False, ranges of F, Cl, Br, I will be set to (0, 0)
        :param parallel: whether to use parallel processing
        :param n_cpu: number of CPU cores used for parallel processing; if -1, all available cores will be used
        :param timeout_secs: timeout in seconds
        :param batch_size: batch size for formula annotation; a larger batch size takes more memory
        :param c_range: C range
        :param h_range: H range
        :param n_range: N range
        :param o_range: O range
        :param p_range: P range
        :param s_range: S range
        :param f_range: F range
        :param cl_range: Cl range
        :param br_range: Br range
        :param i_range: I range
        :param isotope_bin_mztol: m/z tolerance for isotope bin, used for MS1 isotope pattern
        :param max_isotope_cnt: maximum isotope count, used for MS1 isotope pattern
        :param rel_int_denoise_cutoff: relative intensity cutoff, used for MS2 denoise
        :param top_n_per_50_da: top n peaks to keep in each 50 Da, used for MS2 denoise
        """
        if ms_instr is None or ms_instr == "None":
            self.ppm = ppm
            self.ms1_tol = ms1_tol
            self.ms2_tol = ms2_tol
        elif ms_instr in ["orbitrap", "fticr", "qtof"]:
            self.ppm = True
            if ms_instr == "orbitrap":
                self.ms1_tol = 5
                self.ms2_tol = 10
            elif ms_instr == "fticr":
                self.ms1_tol = 2
                self.ms2_tol = 5
            else:
                self.ms1_tol = 10
                self.ms2_tol = 20
        else:
            raise ValueError("Invalid MS instrument. Please choose from 'orbitrap', 'fticr', 'qtof' and None.")

        self.db_mode = 0 if not halogen else 1
        self.parallel = parallel
        if n_cpu > cpu_count() or n_cpu <= 0:
            self.n_cpu = cpu_count()
            if self.parallel:
                logging.info(f"Processing core number is set to {self.n_cpu}.")
        else:
            self.n_cpu = int(n_cpu)

        if timeout_secs <= 0:
            logging.warning("Timeout is set to 300 seconds.")
            self.timeout_secs = 300
        self.timeout_secs = timeout_secs
        if self.parallel:
            self.timeout_secs += 30  # add 30 seconds for db initialization

        if batch_size <= 1:
            self.batch_size = 1000
            logging.warning(f"Batch size is set to {self.batch_size}.")
        else:
            self.batch_size = int(batch_size)

        self.ele_lower = np.array([c_range[0], h_range[0], br_range[0], cl_range[0], f_range[0], i_range[0],
                                   0, n_range[0], 0, o_range[0], p_range[0], s_range[0]], dtype=np.int16)
        self.ele_upper = np.array([c_range[1], h_range[1], br_range[1], cl_range[1], f_range[1], i_range[1],
                                   0, n_range[1], 0, o_range[1], p_range[1], s_range[1]], dtype=np.int16)
        if not halogen:
            self.ele_lower[2:6] = 0
            self.ele_upper[2:6] = 0

        # check valid range
        if np.any(self.ele_lower < 0) or np.any(self.ele_upper < 0):
            raise ValueError("Element range cannot be negative.")
        if np.any(self.ele_lower > self.ele_upper):
            raise ValueError("Element lower bound cannot be larger than upper bound.")

        # check valid param set
        if isotope_bin_mztol <= 0:
            self.isotope_bin_mztol = 0.02
            logging.warning(f"Isotope bin m/z tolerance is set to {self.isotope_bin_mztol}.")
        else:
            self.isotope_bin_mztol = isotope_bin_mztol

        if max_isotope_cnt < 1:
            self.max_isotope_cnt = 4
            logging.warning(f"Maximum isotope count is set to {self.max_isotope_cnt}.")
        else:
            self.max_isotope_cnt = int(max_isotope_cnt)

        if rel_int_denoise_cutoff < 0 or rel_int_denoise_cutoff >= 1:
            self.rel_int_denoise_cutoff = 0.0
            logging.warning(f"Relative intensity denoise cutoff is set to {self.rel_int_denoise_cutoff}.")
        else:
            self.rel_int_denoise_cutoff = rel_int_denoise_cutoff

        if top_n_per_50_da < 1:
            self.top_n_per_50_da = 6
            logging.warning(f"Top n peaks per 50 Da is set to {self.top_n_per_50_da}.")
        else:
            self.top_n_per_50_da = int(top_n_per_50_da)


class Msbuddy:
    """
    msbuddy main class
    msbuddy data is List[MetaFeature]; MetaFeature is a class defined in base/MetaFeature.py
    """
    # singleton pattern
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Msbuddy, cls).__new__(cls)
        return cls._instance

    def __init__(self, config: Union[MsbuddyConfig, None] = None):

        tqdm.write("msbuddy: molecular formula annotation for MS-based small molecule analysis.")

        if config is None:
            self.config = MsbuddyConfig()  # default configuration
        else:
            self.config = config  # customized configuration

        global shared_data_dict  # Declare it as a global variable
        shared_data_dict = init_db()  # database initialization

        self.data = None  # List[MetabolicFeature]

    def update_config(self, **kwargs):
        self.config = MsbuddyConfig(**kwargs)
        global shared_data_dict  # Declare it as a global variable
        shared_data_dict = init_db()  # database initialization

    def load_usi(self, usi_list: Union[str, List[str]],
                 adduct_list: Union[None, str, List[str]] = None):
        self.data = load_usi(usi_list, adduct_list)

    def load_mgf(self, file_path):
        self.data = load_mgf(file_path)

    def add_data(self, data: List[MetaFeature]):
        """
        add customized data
        :param data: metabolic feature list
        """
        self.data = data

    def clear_data(self):
        """
        clear loaded data
        :return: None
        """
        self.data = None

    def _preprocess_and_generate_candidate_formula(self, batch_start_idx: int = 0, batch_end_idx: int = None):
        """
        preprocess data and generate candidate formula space
        :param batch_start_idx: start index of batch
        :param batch_end_idx: end index of batch
        :return: None. Update self.data
        """

        def _preprocess_and_gen_cand_nonparallel(meta_feature: MetaFeature, ps: MsbuddyConfig) -> MetaFeature:
            """
            a wrapper function for data preprocessing and candidate formula space generation
            :param meta_feature: MetaFeature object
            :param ps: Buddy parameter set
            :return: MetaFeature object
            """
            mf = _generate_candidate_formula(meta_feature, ps, shared_data_dict)
            return mf

        batch_data = self.data[batch_start_idx:batch_end_idx]
        modified_mf_ls = []  # modified metabolic feature list, containing annotated results

        # data preprocessing and candidate space generation
        if self.config.parallel:
            with Pool(processes=int(self.config.n_cpu), initializer=_init_pool,
                      initargs=(shared_data_dict,)) as pool:
                async_results = [pool.apply_async(_preprocess_and_gen_cand_parallel,
                                                  (mf, self.config)) for mf in batch_data]
                # Initialize tqdm progress bar
                pbar = tqdm(total=len(batch_data), colour="green", desc="Candidate space generation",
                            file=sys.stdout)
                for i, async_result in enumerate(async_results):
                    pbar.update(1)  # Update tqdm progress bar
                    try:
                        modified_mf = async_result.get(timeout=self.config.timeout_secs)
                        modified_mf_ls.append(modified_mf)
                    except:
                        mf = batch_data[i]
                        logging.warning(f"Timeout for spectrum {mf.identifier}, mz={mf.mz}, rt={mf.rt}, skipped.")
                        modified_mf_ls.append(mf)
            pbar.close()  # Close tqdm progress bar
            del async_results
        else:
            # normal loop, timeout implemented using timeout_decorator
            for mf in tqdm(batch_data, file=sys.stdout, colour="green", desc="Candidate space generation"):
                try:
                    modified_mf = _preprocess_and_gen_cand_nonparallel(mf, self.config)
                    modified_mf_ls.append(modified_mf)
                except:
                    logging.warning(f"Timeout for spectrum {mf.identifier}, mz={mf.mz}, rt={mf.rt}, skipped.")
                    modified_mf_ls.append(mf)

        # update data
        self.data[batch_start_idx:batch_end_idx] = modified_mf_ls
        del modified_mf_ls

    def _assign_subformula_annotation(self, batch_start_idx: int = 0, batch_end_idx: int = None):
        """
        assign subformula annotation for loaded data, no timeout implemented
        :param batch_start_idx: start index of batch
        :param batch_end_idx: end index of batch
        :return: None. Update self.data
        """
        batch_data = self.data[batch_start_idx:batch_end_idx]
        modified_mf_ls = []  # modified metabolic feature list

        if self.config.parallel:
            with Pool(processes=int(self.config.n_cpu)) as pool:
                async_results = [pool.apply_async(_gen_subformula,
                                                  (mf, self.config)) for mf in batch_data]

                pbar = tqdm(total=len(batch_data), colour="green", desc="Subformula assignment: ", file=sys.stdout)
                for i, async_result in enumerate(async_results):
                    pbar.update(1)  # Update tqdm progress bar
                    modified_mf = async_result.get()
                    modified_mf_ls.append(modified_mf)
            pbar.close()  # Close tqdm progress bar
            del async_results
        else:
            # normal loop
            for mf in tqdm(batch_data, desc="Subformula assignment: ", file=sys.stdout, colour="green"):
                modified_mf = _gen_subformula(mf, self.config)
                modified_mf_ls.append(modified_mf)

        # update data
        self.data[batch_start_idx:batch_end_idx] = modified_mf_ls
        del modified_mf_ls

    def annotate_formula(self):
        """
        annotate formula for loaded data
        pipeline: data preprocessing -> formula candidate space generation -> ml model A
        -> subformula annotation -> ml model B -> FDR calculation
        :return: None. Update self.data
        """
        global shared_data_dict  # Declare it as a global variable
        shared_data_dict = init_ml_models(shared_data_dict)  # ml models initialization

        n_batch = self._annotate_formula_prepare()

        # loop over batches
        for n in range(n_batch):
            self._annotate_formula_main_batch(n, n_batch)

        tqdm.write("Job finished.")

    def annotate_formula_cmd(self, output_path: pathlib.Path, write_details: bool = False):
        """
        annotate formula for loaded data, command line version
        write out summary results to a csv file, clear computed data after annotation to save memory
        :param output_path: output path
        :param write_details: whether to write out detailed results
        :return: None
        """
        global shared_data_dict  # Declare it as a global variable
        shared_data_dict = init_ml_models(shared_data_dict)  # ml models initialization

        n_batch = self._annotate_formula_prepare()
        output_path.mkdir(parents=True, exist_ok=True)

        result_summary_df_all = pd.DataFrame()

        # loop over batches
        for n in range(n_batch):
            start_idx, end_idx = self._annotate_formula_main_batch(n, n_batch)
            tqdm.write("Writing batch results...")
            result_summary_df = write_batch_results_cmd(self.data, output_path, write_details,
                                                        start_idx, end_idx)
            result_summary_df_all = pd.concat([result_summary_df_all, result_summary_df], ignore_index=True, axis=0)
            # clear computed data to save memory, convert to None of the same size
            self.data[start_idx:end_idx] = [None] * (end_idx - start_idx)

        tqdm.write("Writing summary results to tsv file...")
        result_summary_df_all.to_csv(output_path / 'msbuddy_result_summary.tsv', sep="\t", index=False)

    def _annotate_formula_prepare(self) -> int:
        """
        prepare for formula annotation
        :return: batch number
        """
        cnt_pre = len(self.data)
        # select MetaFeatures with precursor 1 < mass < 1500
        self.data = [mf for mf in self.data if 1 < mf.mz < 1500]
        cnt_post = len(self.data)
        if cnt_pre != cnt_post:
            tqdm.write(f"{cnt_pre - cnt_post} spectra with precursor mz > 1500 are removed.")

        if not self.data:
            raise ValueError("No data loaded.")
        query_str = f"{len(self.data)} queries loaded." if len(self.data) > 1 else "1 query loaded."
        tqdm.write(query_str)

        if self.config.parallel:
            # parallel processing
            tqdm.write(f"Parallel processing with {self.config.n_cpu} processes.")

        # batches
        n_batch = int(np.ceil(len(self.data) / self.config.batch_size))
        batch_str = f"{n_batch} batches in total." if n_batch > 1 else "1 batch in total."
        tqdm.write(batch_str)

        return n_batch

    def _annotate_formula_main_batch(self, n: int, n_batch: int) -> Tuple[int, int]:
        """
        annotate formula for batch data N
        :param n: batch number
        :param n_batch: total batch number
        :return: start index and end index of batch
        """
        tqdm.write(f"Batch {n + 1}/{n_batch}:")
        # get batch data
        start_idx, end_idx = _get_batch(self.data, self.config.batch_size, n)

        # data preprocessing and candidate space generation
        self._preprocess_and_generate_candidate_formula(start_idx, end_idx)

        # assign subformula annotation
        self._assign_subformula_annotation(start_idx, end_idx)

        tqdm.write("Candidate formula ranking...")
        predict_formula_probability(self.data, start_idx, end_idx, self.config, shared_data_dict)

        # FDR calculation
        calc_fdr(self.data, start_idx, end_idx)

        return start_idx, end_idx

    def get_summary(self) -> List[dict]:
        """
        summarize results
        :return: a list of dictionary containing result summary
        """
        if not self.data:
            raise ValueError("No data loaded.")

        result_summary_list = []
        for mf in self.data:
            result_summary_list.append(mf.summarize_result())

        return result_summary_list

    def mass_to_formula(self, mass: float, mass_tol: float = 10., ppm: bool = True,
                        halogen: bool = False, dbe_cutoff: float = 0.0,
                        integer_dbe: bool = True) -> List[FormulaResult]:
        """
        convert mass to formula, return list of formula strings
        :param mass: target mass, should be <1500 Da
        :param mass_tol: mass tolerance
        :param ppm: whether mass_tol is in ppm
        :param halogen: whether to include halogen atoms
        :param dbe_cutoff: double bond equivalent cutoff
        :param integer_dbe: whether to return only integer DBE
        :return: list of FormulaResult objects
        """
        formulas = query_neutral_mass(mass, mass_tol, ppm, halogen, shared_data_dict)

        if integer_dbe:
            out = [FormulaResult(form_arr_to_str(f.array), f.mass, mass) for f in formulas
                   if f.dbe >= dbe_cutoff and f.dbe % 1 == 0 and senior_rules(f.array)]
        else:
            out = [FormulaResult(form_arr_to_str(f.array), f.mass, mass) for f in formulas
                   if f.dbe >= dbe_cutoff and senior_rules(f.array)]

        # sort by absolute mass error, ascending
        out.sort(key=lambda x: abs(x.mass_error))

        return out

    def mz_to_formula(self, mz: float, adduct: str = '[M+H]+',
                      mz_tol: float = 10., ppm: bool = True, halogen: bool = False,
                      dbe_cutoff: float = 0.0, integer_dbe: bool = True) -> List[FormulaResult]:
        """
        convert mz to formula, return list of formula strings
        :param mz: target mz, should be <1500 Da
        :param adduct: adduct string
        :param mz_tol: mz tolerance
        :param ppm: whether mz_tol is in ppm
        :param halogen: whether to include halogen atoms
        :param dbe_cutoff: double bond equivalent cutoff
        :param integer_dbe: whether to return only integer DBE
        :return: list of FormulaResult objects
        """
        valid_adduct, pos_mode = check_adduct(adduct)
        if not valid_adduct:
            raise ValueError("Invalid adduct string.")

        # query
        ion = Adduct(adduct, pos_mode)

        if halogen:
            formulas, _ = query_precursor_mass(mz, ion, mz_tol, ppm, 1, shared_data_dict)
        else:
            formulas, _ = query_precursor_mass(mz, ion, mz_tol, ppm, 0, shared_data_dict)

        ion_int = 1 if pos_mode else -1

        if integer_dbe:
            out = [FormulaResult(form_arr_to_str(f.array),
                                 (f.mass * ion.m + ion.net_formula.mass - ion_int * 0.00054858) / abs(ion.charge),
                                 mz) for f in formulas if f.dbe >= dbe_cutoff and f.dbe % 1 == 0 and senior_rules(f.array)]
        else:
            out = [FormulaResult(form_arr_to_str(f.array),
                                 (f.mass * ion.m + ion.net_formula.mass - ion_int * 0.00054858) / abs(ion.charge),
                                 mz) for f in formulas if f.dbe >= dbe_cutoff and senior_rules(f.array)]

        # sort by absolute mass error, ascending
        out.sort(key=lambda x: abs(x.mass_error))
        return out


def _get_batch(data: List[MetaFeature], batch_size: int, n: int):
    """
    get batch data
    :param data: data list
    :param batch_size: batch size
    :param n: batch number
    :return: batch data
    """
    start_idx = n * batch_size
    end_idx = min((n + 1) * batch_size, len(data))
    return start_idx, end_idx


def _init_pool(the_dict):
    """
    initialize pool for parallel processing
    :param the_dict: global dictionary containing shared data
    :return: None
    """
    global shared_data_dict
    shared_data_dict = the_dict


def _preprocess_and_gen_cand_parallel(meta_feature: MetaFeature, ps: MsbuddyConfig) -> MetaFeature:
    """
    a wrapper function for data preprocessing and candidate formula space generation
    :param meta_feature: MetaFeature object
    :param ps: MsbuddyConfig object
    :return: MetaFeature object
    """
    mf = _generate_candidate_formula(meta_feature, ps, shared_data_dict)
    return mf


def _gen_subformula(mf: MetaFeature, ps: MsbuddyConfig) -> MetaFeature:
    """
    a wrapper function for subformula generation
    :param mf: MetaFeature object
    :param ps: MsbuddyConfig object
    :return: MetaFeature object
    """
    if not mf.ms2_processed:
        return mf

    if not mf.candidate_formula_list:
        return mf

    # assign subformula annotation
    mf = assign_subformula_cand_form(mf, ps.ppm, ps.ms2_tol)

    # retain candidate formula with subformula annotations if there is any candidate formula with annotations
    cand_form_exp_ms2_peak_list = []
    for cf in mf.candidate_formula_list:
        if cf.ms2_raw_explanation:
            cand_form_exp_ms2_peak_list.append(len(cf.ms2_raw_explanation))
        else:
            cand_form_exp_ms2_peak_list.append(0)
    if min(cand_form_exp_ms2_peak_list) > 0:
        mf.candidate_formula_list = [cf for cf in mf.candidate_formula_list if len(cf.ms2_raw_explanation) > 0]

    return mf


def _generate_candidate_formula(mf: MetaFeature, ps: MsbuddyConfig, global_dict) -> MetaFeature:
    """
    preprocess data and generate candidate formula space
    :param mf: MetaFeature object
    :param ps: MsbuddyConfig object
    :param global_dict: global dictionary containing shared data
    :return: MetaFeature object
    """
    # data preprocessing
    mf.data_preprocess(ps.ppm, ps.ms1_tol, ps.ms2_tol, ps.isotope_bin_mztol, ps.max_isotope_cnt,
                       ps.rel_int_denoise_cutoff, ps.top_n_per_50_da)
    # generate candidate formula space
    gen_candidate_formula(mf, ps.ppm, ps.ms1_tol, ps.ms2_tol, ps.db_mode, ps.ele_lower, ps.ele_upper,
                          ps.max_isotope_cnt, global_dict)
    return mf


if __name__ == '__main__':
    import time

    start = time.time()

    msb_config = MsbuddyConfig(ms_instr='orbitrap',  # supported: "qtof", "orbitrap" and "fticr"
                               halogen=True, parallel=True, n_cpu=10)

    msb_engine = Msbuddy(msb_config)

    msb_engine.load_mgf('/Users/shipei/Documents/projects/msbuddy/demo/input_file.mgf')

    # mgf_folder = '/Users/shipei/Documents/projects/msbuddy/results/lcms_datasets/MSV000085143_chagas_neg_orbi'  # 83
    # mgf_folder = '/Users/shipei/Documents/projects/msbuddy/results/lcms_datasets/MSV000081463_tomato_pos'  # 194
    # mgf_folder = '/Users/shipei/Documents/projects/msbuddy/results/lcms_datasets/MSV000086988_fecal_neg_orbi'  # 118
    # mgf_folder = '/Users/shipei/Documents/projects/msbuddy/results/lcms_datasets/MSV000081981_AmericanGutProject'  # 146
    # msb_engine.load_mgf(str(mgf_folder + '/ms1_ms2.mgf'))

    # cmd version
    # msb_engine.annotate_formula_cmd(pathlib.Path(str(mgf_folder + '/msbuddy_output')), True)

    # annotate molecular formula
    msb_engine.annotate_formula()

    end = time.time()
    print(end - start)

    # retrieve the annotation result summary
    result = msb_engine.get_summary()

    print(result)
    form_top1 = [r['formula_rank_1'] for r in result]
    form_est_fdr = [r['estimated_fdr'] for r in result]
    print(form_top1)
    print(form_est_fdr)
